#ifndef _RDBINGRESSQLDA_H_
#define _RDBINGRESSQLDA_H_

// This class is the keeper of the backend's sqlda data structure.
// It makes sure that the sqlda is used properly and does conversions.

// Should this class be generalized?

#include "RdbItem.h"
#include "Boolean.h"
#include "bitvector.h"

#ifndef _SQLDA_H_
#define _SQLDA_H_
#include <eqsqlda.h>
#endif

class RdbIngresSqlda
{
  public:

    RdbIngresSqlda(int items = 0);
				// allocate room for 'items' columns 

    ~RdbIngresSqlda();

    void Prepare();		// get storage for items described in sqlda

    void Prepare(const RdbIngresSqlda& other, const BitVector& whichCols);
				// set up this sqlda like other,
				// but only the items in whichCols are used 

    void Unprepare();		// release all of the prepared elements

    int Arity() const;		// number of elements prepared

    int Size() const;		// max number of elements that can be prepared

    void Resize(int items);	// adjust max; ptrs to sqlda invalidated!

    void SetItem(int i, const RdbItem& val);
				// set the i'th item to val

    Boolean IsNull(int i) const; // is the i'th item null?

    RdbItem::ItemType Type(int i) const;
				// what is the data type of col i

    RdbItem operator [](int i) const;
				// gimme item i

    operator IISQLDA*() const;	// return a pointer to the sqlda

  private:

    IISQLDA* sqlda;		// the sqlda to communicate with the dbms

    int itemsPrepared;		// how many items are allocated in the sqlda

    RdbIngresSqlda(const RdbIngresSqlda&); // na

    RdbIngresSqlda& operator =(const RdbIngresSqlda&); // na
};


#include "RdbIngresSqlda.i"

#endif // _RDBINGRESSQLDA_H_
