#include <iostream.h>
#include <string.h>
#include "List.h"


PtrListBase::PtrListBase( int startSize )
{
    allocated = 0;
    count = 0;
    list = NULL;
    Resize(startSize);
}


PtrListBase::PtrListBase( const PtrListBase& other )
{
    allocated = 0;
    count = 0;
    list = NULL;
    *this = other;
}


PtrListBase::~PtrListBase()
{
    DeleteAll();
    free(list);
    list = NULL;
    allocated = count = 0;
}


PtrListBase& PtrListBase::operator =(const PtrListBase& other)
{
    if( this != &other ) {	// a = a
	DeleteAll();
	Resize(other.Size());
	count = other.Size();
	memcpy( list, other.list, sizeof(void*) * count );
    }
    return *this;
}
    

void PtrListBase::Resize(int s)
{
    if( s > allocated ) {	// need more than allocated
	allocated *= 2;		// double size of array
	if( s > allocated )	// need more than twice!
	  allocated = s;	// just get enough
	if( list == NULL )	// should not need this, but realloc 
				// is not working on a null list
	  list = (void**) malloc( sizeof(void*) * allocated );
	else
	  list = (void**) realloc( list, sizeof(void*) * allocated );
	memset(list + Size(), 0, (allocated - Size()) * sizeof(void*));
				// initialize all pointers to null
    }
}


int PtrListBase::Append(void* element, int n)
{
    Resize(count+n);
    int old_count = count;
    count += n;
    for( int i = old_count ; i < count ; i++ ) {
	list[i] = element;
    }
    return old_count;
}


int PtrListBase::Insert( void* element, int loc )
{
    if( loc >= Size() )
      Resize(loc+1);
    Resize(count+1);
    for( int i = count ; i > loc ; i-- )
      list[i] = list[i-1];
    list[loc] = element;
    count++;
    return loc;
}


void* PtrListBase::Delete( int i )
{
    assert( ValidSubscript(i) );
    void* result = list[i];
    for( i++ ; i < count ; i++ )
      list[i-1] = list[i];
    count--;
    return result;
}

