// *************************************************************************
//    BuiltinRelation.h
// *************************************************************************

#ifndef _BuiltinRelation_H_
#define _BuiltinRelation_H_

#include "Relation.h"
#include "Debug.h"

// *************************************************************************

class Cor_BuiltinRelation;
class Cor_TupleIterator;

typedef Cor_BindEnv* (*SolveProc)(Cor_BuiltinRelation&, Cor_TupleIterator&);
typedef Cor_Tuple* (*UserTupleProc)(Cor_Tuple  *query);
typedef Cor_Relation* (*UserRelProc)(Cor_Tuple *query);

// *************************************************************************
//    Each builtin-relation is associated with a 'solver'
//    function that is invoked when a get_next() occurs it.

class Cor_BuiltinRelation : public Cor_Relation {

  public:
    SolveProc solver;

    Cor_BuiltinRelation(int a, Cor_CArg *n, SolveProc solve_proc, int insert=1);
    Cor_BuiltinRelation(char *n, int a, UserRelProc solve_proc);
    Cor_BuiltinRelation(char *n, int a, UserTupleProc solve_proc);

    virtual Cor_BindEnv *get_next(Cor_TupleIterator&);
    virtual void release(Cor_TupleIterator&);
    virtual Cor_Tuple *get_next_tuple(Cor_TupleIterator& it) {
        get_next(it);
        return(NULL);
    }

    virtual int insert_tuple(Cor_Tuple *tuple);
    virtual int insert_new(Cor_ArgList &args, Cor_BindEnv *env);

    virtual void print_name(ostream *file);
    virtual int is_printable_builtin() {return 1;}
};

// *************************************************************************

#endif _BuiltinRelation_H_
