// *************************************************************************
//    BuiltinUtil.C
// *************************************************************************

#include <stdio.h>

#include "Util.h"
#include "Globals.h"
#include "CoralError.h"
#include "Debug.h"

// *************************************************************************
int set_out_strm(char *file_name)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (   (Cor_exEnv->out_strm != cout_strm)
        && (Cor_exEnv->out_strm != cerr_strm)
        && (Cor_exEnv->out_strm != Cor_exEnv->err_strm)
        && (Cor_exEnv->out_strm != Cor_exEnv->trace_strm)
        && (strcmp(Cor_exEnv->out_filename, Cor_exEnv->trace_filename)))
    {
        delete Cor_exEnv->out_strm;
        delete [] Cor_exEnv->out_filename;
    }
        
    if (!strcmp(file_name, "stdout")) {
        Cor_exEnv->out_strm = cout_strm;
        Cor_exEnv->out_filename = new char[7];
        strcpy(Cor_exEnv->out_filename, "stdout");
    }

    else if (!strcmp(file_name, "stderr")) {
        Cor_exEnv->out_strm = cerr_strm;
        Cor_exEnv->out_filename = new char[7];
        strcpy(Cor_exEnv->out_filename, "stderr");
    }

    else if (!strcmp(file_name, Cor_exEnv->trace_filename)) {
        Cor_exEnv->out_strm = Cor_exEnv->trace_strm;
        Cor_exEnv->out_filename = Cor_exEnv->trace_filename;
    }
    else if ((Cor_exEnv->out_strm = Cor_open_outfile(file_name,ios::app)))
        Cor_exEnv->out_filename = file_name;
    else {
        Cor_exEnv->out_strm = cout_strm;
        Cor_exEnv->out_filename = new char[7];
        strcpy(Cor_exEnv->out_filename, "stdout");
        CORAL_error(COR_CANNOT_OPEN_FILE, file_name, "set_out_strm",
            __FILE__, __LINE__);
        return 0;
    }
    return 1;
}

// *************************************************************************
int set_trace_strm(char *file_name)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (   (Cor_exEnv->trace_strm != cout_strm)
        && (Cor_exEnv->trace_strm != cerr_strm)
        && (Cor_exEnv->trace_strm != Cor_exEnv->out_strm)
        && (Cor_exEnv->trace_strm != Cor_exEnv->err_strm)
        && (strcmp(Cor_exEnv->out_filename, Cor_exEnv->trace_filename)))
    {
        delete Cor_exEnv->trace_strm;
        delete [] Cor_exEnv->trace_filename;
    }
    
    if (!strcmp(file_name, "stdout")) {
        Cor_exEnv->out_strm = cout_strm;
        Cor_exEnv->out_filename = new char[7];
        strcpy(Cor_exEnv->out_filename, "stdout");
    }
    else if (!strcmp(file_name, "stderr")) {
        Cor_exEnv->out_strm = cerr_strm;
        Cor_exEnv->out_filename = new char[7];
        strcpy(Cor_exEnv->out_filename, "stderr");
    }
    else if (!strcmp(file_name, Cor_exEnv->error_filename)) {
        Cor_exEnv->trace_strm = Cor_exEnv->err_strm;
        Cor_exEnv->trace_filename =
            new char[strlen(Cor_exEnv->error_filename)+1];
        strcpy(Cor_exEnv->trace_filename, Cor_exEnv->error_filename);
    }
    else if (!strcmp(file_name, Cor_exEnv->out_filename)) {
        Cor_exEnv->trace_strm = Cor_exEnv->out_strm;
        Cor_exEnv->trace_filename = Cor_exEnv->out_filename;
    }
    else if ((Cor_exEnv->trace_strm = Cor_open_outfile(file_name,ios::app)))
        Cor_exEnv->trace_filename = file_name ;
    else {
        Cor_exEnv->trace_strm = Cor_exEnv->err_strm;
        Cor_exEnv->trace_filename = new char[7];
        strcpy(Cor_exEnv->trace_filename, "stderr");
        CORAL_error(COR_CANNOT_OPEN_FILE, file_name, "set_trace_strm",
            __FILE__, __LINE__ );
        return 0;
    }
    return 1;
}

// *************************************************************************
//  Executes the magic program using the system() call. The input file
//  to magic is <infile_name>, and the output is <outfile_name>.M
// *************************************************************************
int Cor_callMagic(char *infile_name, char *outfile_name)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char term_str[400];

    sprintf(term_str, "%s/bin/magic -i %s > %s.M",
            Cor_exEnv->coral_path, infile_name, outfile_name);

    if (system(term_str))
        return(0);
    
    return(1);
}

// *************************************************************************
int Cor_tableFile(const char *name)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int l = strlen(name);

    if ((name[l-2] == '.')
        && ((name[l-1] == 'T') || (name[l-1] == 'F'))) {
        // '.T' specifies a table file, '.F' is a fact file
        return(1);
    }
    return(0);
}

// *************************************************************************
//    Files ending in .S, .C and .o are to be incrementally loaded
int Cor_incrLoadFile(const char *name)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int l = strlen(name);

    if ((name[l-2] == '.')
            && ((name[l-1] == 'S')
                || (name[l-1] == 'C')
                || (name[l-1] == 'o'))) {
        return(1);
    }
    return(0);
}

// *************************************************************************
