// *************************************************************************
//    PersistentIndex.C
// *************************************************************************

#ifdef WITH_PERSISTENCE


#include "PersistentRelation.h"


#include "Globals.h"
#include "CoralError.h"
#include "SMUtil.h"
#include "sm_client.h"
#include "Config.h"
#include "Schema.h"
#include "IndexStruct.h"
#include "BPlusTreeIndex.h"
#include "FileDescStruct.h"
#include "BufferGroup.h"
#include "Transaction.h"
#include "Catalog.h"
#include "StorageManager.h"

#include "PersistentIndex.h"
#include "Debug.h"

// *************************************************************************

extern int Cor_unify_literal(Cor_TupleIterator&, Cor_Tuple *, int) ;
extern int SM_initialized;

// *************************************************************************

#define DELETE_INDEX_ENTRY 1
#define DELETE_OBJ_AND_INDEX_ENTRY 2

// temporary hack to avoid changing generic-index.h
// also see persistent-index.C  :: PRAVEEN
#define COR_I_PARTIAL_SUCCESS  2

// *************************************************************************
/*
    Constructor that initializes the persistent-index structure from the
    catalog if it exists in the catalog.  Otherwise, an error is raised.

    Note that the file names are only guarenteed to be unique upto
    COR_MAX_FILE_NAME_LEN
*/

Cor_PersistentIndex::Cor_PersistentIndex(Cor_PersistentRel *rel,
    Cor_BitVector *bv)
    : Cor_GenericIndex((Cor_StorageRelation *)rel, bv)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
}

// *************************************************************************
Cor_Tuple *Cor_PersistentIndex::get_next_tuple (Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
  if (!SM_initialized) return NULL;

  Cor_Tuple *tuple;

  return p_rel()->indexScanGetNext(iterator, chosen_index);

}

// *************************************************************************
Cor_BindEnv *Cor_PersistentIndex::get_next(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized) return NULL;

    Cor_Tuple *tuple = get_next_tuple(iterator);
    if (iterator.no_match())
        return NULL;
    else
        return iterator.bindenv;
}

// *************************************************************************

#endif WITH_PERSISTENCE

