
// 
//   CallStk.h  - keep track of the call stack (for debugging)
// 
//   Shaun Flisakowski, 1996

#ifndef   _CALL_STK_H
#define   _CALL_STK_H

// Only keeps track of these things when DO_CALL_STACK is set.


#include <iostream.h>

    // This implementation uses a global CallStk variable.
class CallStk;
extern CallStk  *callstk;

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

enum Entry_Type {
    Function_UnUsed,    // An unused slot
    Function_Entry,
    Function_Exit
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

class CS_StrTab {

  private:
    char        **data;       // Array of pointers to strings
    unsigned int  num;        // The number of strings stored
    unsigned int  length;     // The length of the array

  public:

    CS_StrTab();
   ~CS_StrTab();

      // returns index in the table
    unsigned int add(const char *str);

      // get the string pointed to by index (stays in table)
    const char *get(unsigned int index);
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

class CallStk_Item;

  // This class goes in and out of scope at function entry/exit
class CallStk_Mark {
    CallStk_Item    *item;    // Needed on exit
  public:
    CallStk_Mark( char *func, char *file, int lineno );
   ~CallStk_Mark();
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class CallStk_Item {
  friend class CallStk;

  private:
    unsigned int func;    // This function's name
    unsigned int file;    // The filename
    int          lineno;  // The line number of the entry pt

    int          active;  // We are active if we have not
                          // left this function yet.

  public:
    CallStk_Item( char *_func, char *_file, int _lineno );
   ~CallStk_Item();

    void show_call_item(ostream *out);
    friend ostream& operator <<(ostream& out, const CallStk_Item& itm);
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class CallStk_Entry {
  friend class CallStk;

  private:
    Entry_Type     type;    // entry, exit, or unused?
    CallStk_Item  *item;    // The info (shared by entry/exit)

  public:
    CallStk_Entry();
   ~CallStk_Entry();

    CallStk_Entry& operator =(const CallStk_Entry& src);

    friend ostream& operator <<(ostream& out, const CallStk_Entry& itm);
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class CallStk {
  friend class CallStk_Item;
  friend class CallStk_Entry;

  private:
    int              num;  // Number of entries in the stack
    int              len;  // Size of allocated memory

    CallStk_Entry   *stk;  // The array of stack entries

        // Prepare the stk to recieve a new item.
    void checksize();

      // Squash the stack, removing unused entries
    void compress();

  public:
    CS_StrTab filenames;
    CS_StrTab funcnames;

    CallStk();
   ~CallStk();

    void entry(CallStk_Item *item);
    void exit(CallStk_Item *item);

        // Throw away the stack trace for all inactive entries
    void reset();

        // Show the call stack, to depth given (0 == "all")
    void show_call_stack(ostream* out, int depth=0);

        // Show all calls in the trace
    void show_call_trace(ostream *out);
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

#endif
