// *************************************************************************
//    QFPipelinedModuleData.C
// *************************************************************************

#include "QFPipelinedModuleData.h"

#include "ParserStruct.h"
#include "Globals.h"
#include "Relation.h"
#include "PipelinedDerivedRelation.h"
#include "Debug.h"

// *************************************************************************
Cor_QFPipelinedModuleData::Cor_QFPipelinedModuleData (Cor_Table *table,
    Cor_RuleInfoArray *rArray, Cor_ParserStruct *Cor_parserStruct)
    : backtrackInfo(rArray)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    F = table;
    ruleArray = rArray;

      // Need to set the parserStruct's pointers to NULL
      // so that these items are not deleted
    multiset_annotations = Cor_parserStruct->MultisetAnnotations;
    Cor_parserStruct->MultisetAnnotations    = NULL;
    index_annotations = Cor_parserStruct->make_index_annotations;
    Cor_parserStruct->make_index_annotations = NULL;
    agg_sel_annotations = NULL;
    diskrel_annotations = NULL;
    asyncrel_annotations = NULL;
    
    TableSetUp = 0;            // no pipelined execs in progress
    rel_options = 0;
    predicates = new Cor_ModulePred(NULL);

    indices = NULL;
    maxNumVars = NULL;
    maxNumLiterals = NULL;

    pred_ptrs = NULL;
}

// *************************************************************************
// Clean the Local Pipelined Relations out of the Table.
// *************************************************************************
static void clean_table(Cor_Table *F)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
      // Delete Local Pipelined relations that would otherwise be leaked
    int j;
    for (j = 0; j < F->numRels; j++) {
        if (F->rels[j].relPtr && F->rels[j].relPtr->is_localpiped()) {
            
            Cor_PipelinedDerivedRelation *del_rel =
                (Cor_PipelinedDerivedRelation *) F->rels[j].relPtr;

              // Need to avoid a circular relationship
            F->rels[j].ClearRel();
            del_rel->md = NULL;

            del_rel->empty_relation(1);
            delete del_rel;
        }
    }
}

// *************************************************************************
Cor_QFPipelinedModuleData::~Cor_QFPipelinedModuleData ()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_ModulePred *curr, *prev = NULL;
    
    clean_table(F);

    for (curr = predicates; curr; curr=curr->next) {
        delete prev;
        prev = curr;
    }
    delete prev;

    for (int j=0; j < ruleArray->num_rules(); j++)
        delete [] indices[j];

    delete ruleArray;

    delete [] indices;
    delete [] maxNumVars;
    delete [] maxNumLiterals;

    delete [] pred_ptrs;
}

// *************************************************************************
void Cor_QFPipelinedModuleData::clean_cache()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    for (int i = 0; i< ruleArray->num_rules();i++) {
        for (int j = 0;j< (ruleArray->rInfoPtrs[i])->num_literals; j++) {
            indices[i][j] = -1;
        }
    }
}

// *************************************************************************
