// *************************************************************************
//    DerivedRelation.C
// *************************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "DerivedRelation.h"

#include "TupleIterator.h"
#include "Globals.h"
#include "Unparse.h"
#include "QFMaterializedModuleData.h"
#include "Execute.h"
#include "CoralError.h"
#include "GenericIndex.h"
#include "HashSimpleRelation.h"
#include "ParseActions.h"
#include "ParseUtils.h"
#include "ParserEnv.h"
#include "ParserStruct.h"
#include "CoralBreak.h"
#include "Util.h"
#include "Debug.h"

#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/param.h>

// *************************************************************************

char *Cor_my_str;

// *************************************************************************
Cor_DerivedRelation::Cor_DerivedRelation(Cor_CArg& _name, int new_arity)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    name = _name;
    _arity = new_arity;
    methods = NULL;
    has_release = 1;
    r_kind = COR_R_DERIVED;
}

// *************************************************************************
Cor_DerivedRelation::~Cor_DerivedRelation()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_DerivedMethod *meth = methods, *next_meth;
    while (meth) {
        next_meth = meth->next;
        delete meth;
        meth = next_meth;
    }
}

// *************************************************************************
Cor_BindEnv *Cor_DerivedRelation::get_next(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    this->get_next_tuple(iterator);
    if (iterator.no_match())
        return NULL;
    return(iterator.bindenv);
}

// *************************************************************************
Cor_Tuple *Cor_DerivedRelation::get_next_tuple(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#    ifdef DO_PROFILE
    if (Cor_T_Stack->count > 0)
      (Cor_T_Stack->cur_md()->module_info->Cor_GlobalProfileInfo.getnexts)++;
    else
      (Cor_GlobalProfileInfo.getnexts)++;
#    endif

    int i;
    Cor_StorageRelation *tuples;
    Cor_TupleIterator   *tuples_iterator;

    if (iterator.ipos++ == 0) {
        Cor_BindEnv *env = iterator.bindenv;
        iterator.tuple_env = NULL;
    
        int arity = iterator.arg_list.count();
        // Find adornment of query.
        Cor_BitVector adornment(arity);

        for (i = 0; i < arity; i++) {
            Cor_Term term1(iterator.arg_list[i], iterator.bindenv);
            Cor_full_dereference_term(term1);

            if (term1.expr.simplify_me(iterator.bindenv))
                adornment.set(i,1);
            else
                adornment.set(i,0);
        }

        // Find a method that matches the query's adornment.

        Cor_DerivedMethod *method;

        // Best method with insufficent 'b' (where query is 'f').
        Cor_DerivedMethod *best_underbound_method = NULL;

        // Count of 'b's short in best_underbound_method.
        int best_underbound_shortfall = arity + 1;

        // Best method with excess 'b' (where query is 'f').
        Cor_DerivedMethod *best_overbound_method = NULL;

        int least_overbound_excess = arity + 1;
        
        for (method = methods; ; method = method->next) {
            int adorn_delta[3];
            if (method == NULL) {
                if ((method = best_underbound_method) != NULL) {
                    if (best_underbound_shortfall &&
                        !Cor_exEnv->C_quiet_mode_default)
                    {
                        *(Cor_exEnv->err_strm) << 
                            "CORAL::Warning : Using underbound method ";
                        (method->adornment).printon(Cor_exEnv->err_strm);
                        *(Cor_exEnv->err_strm) << " for "
                            << name.symbol() << " ";
                        adornment.printon(Cor_exEnv->err_strm);
                        *(Cor_exEnv->err_strm) << "!\n";
                    }
                    break;
                }
                if ((method = best_overbound_method) != NULL) {
                    if (least_overbound_excess
                            && !Cor_exEnv->C_quiet_mode_default) {
                        *(Cor_exEnv->err_strm) << 
                            "CORAL::Warning : Using overbound method ";
                        (method->adornment).printon(Cor_exEnv->err_strm);
                        *(Cor_exEnv->err_strm) << " for "
                            << name.symbol() << " ";
                        adornment.printon(Cor_exEnv->err_strm);
                        *(Cor_exEnv->err_strm) << endl;
                        if (method->md->module_info->UseOrdSearch) {
                            *(Cor_exEnv->err_strm)
                                << "CORAL:: Cannot use Ordered Search"
                                    " with overbound method. Sorry.\n";
                            iterator.set_no_match();
                            return(NULL);
                        }
                    }
                    break;
                }

                *(Cor_exEnv->err_strm)
                    << "No rule matches the query " << name.symbol()
                    << " (with adornment ";
                Cor_print_adornment(adornment, Cor_exEnv->err_strm);
                *(Cor_exEnv->err_strm) << ")\n";
                iterator.set_no_match();
                return(NULL);
            }

            if (!method->valid_method)
                continue;

            int i = adornment.len();
            if (i != method->adornment.len())
                continue;

            // Use adorn_delta to summarizes the difference between
            // adornment (from the query) and method->adornment.
            // adorn_delta[0] counts bits where adornment is 'f' (0)
            // and method.adornment is 'b' (1).
            // adorn_delta[1] counts bits where the adornments are equal.
            // adorn_delta[2] counts bits where adornment is 'b'
            //   and method.adornment is 'f'.
            adorn_delta[0] = 0;
            adorn_delta[1] = 0;
            adorn_delta[2] = 0;
            while (--i >= 0) {
                adorn_delta[1+adornment.test(i)-method->adornment.test(i)]++;
            }

            if (adorn_delta[0] + adorn_delta[2] == 0) // Exact match.
                break;

            if (adorn_delta[0] == 0 &&
                best_underbound_shortfall > adorn_delta[2])
            {
                // adornment.implies(method->adornment)
                best_underbound_method = method;
                best_underbound_shortfall = adorn_delta[2];
            }
            else if (least_overbound_excess > adorn_delta[0]) {
                // adornment.implies(method->adornment)
                best_overbound_method = method;
                least_overbound_excess = adorn_delta[0];
            }
        }

        Cor_my_str = method->md->export_pred->name.symbol();
        Cor_opt = (method->md->rel_options | Cor_exEnv->GlobalRelationOptions)
            & REL_EXPLAIN;

        
#        ifdef DO_EXPLAIN
        if ((method->md->rel_options | Cor_exEnv->GlobalRelationOptions) &
                REL_EXPLAIN)
        {
            int childpid;
            char fifo_path[MAXPATHLEN];

            if ((Cor_exEnv->explain_strm == NULL)
                        && (! coralConfig->server_mode)){

                sprintf(fifo_path,"/tmp/explain_fifo");
                    // Remove the old FIFO if nessecary
                unlink(fifo_path);

                if ( mkfifo(fifo_path, S_IRUSR | S_IWUSR ) < 0)
                    *(Cor_exEnv->err_strm)
                          << "Can't create FIFO for explain.\n";


                if ((childpid = fork()) < 0) {
                       // fork another process to run explain
                     *(Cor_exEnv->err_strm) << "can't fork\n";
                }
                else if (childpid > 0) {

                    // parent process
                    Cor_my_str =method->md->export_pred->name.symbol();

                    if (!(Cor_exEnv->explain_strm
                                        = Cor_open_outfile(fifo_path))) {
                        CORAL_error(COR_CANNOT_OPEN_FILE,
                            "file_name" , "explain");
                        return(NULL);
                    }

                    *(Cor_exEnv->explain_strm) << "coral.\n";
                    *(Cor_exEnv->explain_strm)
                        << "module_name(\"" << Cor_my_str
                        << "\",0).\n" << flush;

                } else {
                    int fd;
                    char *path = new char[MAXPATHLEN];
                    char *path2 = new char[MAXPATHLEN];
                    if ((fd = open(fifo_path, O_RDONLY)) < 0)
                        *(Cor_exEnv->err_strm) 
                            << "Cannot open " << fifo_path << endl; 

                    dup2(fd,0); // duplicate file descriptor

                    sprintf(path,"%s/bin/coralsh",Cor_exEnv->coral_path);
                    sprintf(path2,"%s/bin/Explain",Cor_exEnv->coral_path);

                    execl(path, "coralsh", "-f", path2, (char*) 0);

                        /* NOTREACHED */
                    delete [] path;
                    delete [] path2;
                    exit(0);
                }

            } else {

                if (Cor_exEnv->explain_strm == NULL){
                    *(Cor_exEnv->err_strm) 
                        << "Explain not connected - turning explain off.\n"
                        << flush;
                    Cor_exEnv->GlobalRelationOptions &= ~REL_EXPLAIN;
                } else {

                      // Already have a FIFO (Explain is running) 
                    Cor_my_str= method->md->export_pred->name.symbol();
                    *(Cor_exEnv->explain_strm) << "module_name(\""
                            << Cor_my_str << "\",0).\n" << flush;
                }
            }
        }

#        endif    // DO_EXPLAIN

        iterator.moduledata = method->md;

        Cor_StackMark stack_mark;
        
        // Insert the magic tuple.
        int magic_arity = method->adornment.count();
        Cor_ArgList *magic_args = new Cor_ArgList(magic_arity);
        int arg_bit = -1;
        for (i = 0; i < magic_arity; i++) {
            arg_bit = method->adornment.next(arg_bit);
            (*magic_args)[i] = iterator.arg_list[arg_bit];
        }
      
        Cor_StorageRelation *magic_rel = method->md->magic_rel;
        delete magic_rel;
        magic_rel = method->md->magic_rel = NULL;
        int old_magic_rel = 0;
        
        {
            char magic_name[200];
            Cor_sprint_magic_name(magic_name, name.symbol(),
                &(method->adornment));

            Cor_CArg magic_arg = Cor_CArg(magic_name);
            magic_rel = Cor_make_local_relation(magic_arg,
                magic_arity, method->md->module_info->IndexDeltas);
            magic_rel->local_options = method->md->rel_options;
        }

        magic_rel->insert_new(*magic_args, iterator.bindenv);
        delete magic_args;
 
        // Solve the query.
        tuples = (*method->interp_solver)(method->md, magic_rel,
                old_magic_rel, &iterator);

        if (!tuples) {
            *(Cor_exEnv->err_strm) << "Error: Cannot solve query.\n";
            iterator.set_no_match();
            return(NULL);
        }
        
        stack_mark.pop_to();
        
        Cor_RMark *endmark = (Cor_RMark *) NULL;

        tuples_iterator =
          new Cor_TupleIterator(tuples, iterator.arg_list, env, NULL, endmark);

        iterator.answer_iterator = tuples_iterator;
        tuples_iterator->state = iterator.state;

        delete method->md->magic_rel;
        method->md->magic_rel = magic_rel;
    } else {
        tuples_iterator = iterator.answer_iterator;
        tuples = (Cor_StorageRelation *)(tuples_iterator->relation);
    }
    
    Cor_Tuple *tuple = tuples->get_next_tuple(*tuples_iterator);
    
    // Careful here - we are returning a different tuple iterator, and
    // must make sure all necessary arguments are copied over as required.

    while (tuples_iterator->no_match() && iterator.state) {
        iterator.reset_no_match();
        (*(iterator.state)->solver)((iterator.state)->md,
            (iterator.state)->magic, 0, &iterator);
        tuple = tuples->get_next_tuple(*tuples_iterator);
    }

    if (tuples_iterator->no_match()) {
        iterator.set_no_match();
        tuple = NULL;
    } else {
        iterator.reset_no_match();
        iterator.tuple_env = tuples_iterator->tuple_env;
        tuples_iterator->tuple_env = NULL;
    }
   
    return(tuple);
}


// *************************************************************************
/*
 * WARNING:: This function may be called multiple times on an iterator.
 * It must make sure things are not deleted multiple times.
 */

void Cor_DerivedRelation::release(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (iterator.answer_iterator)  {
        Cor_TupleIterator *tuples_iterator = iterator.answer_iterator;
        Cor_StorageRelation *tuples =
                (Cor_StorageRelation *) tuples_iterator->relation;

        if (iterator.state) {
            // in the case of lazy evaluation, it is possible that
            // there may still be some saved state. this happens when
            // a cut is used to suddenly terminate a computation

            if (iterator.state->rm_Arr)
                delete iterator.state->rm_Arr;

            Cor_cleanup_do_query(iterator.state->md,
                    iterator.state->magic, iterator.state->table);
            delete iterator.state;
            iterator.state = NULL;
        }
        
        Cor_reclaim_relation(tuples);
        tuples_iterator->relation = NULL;

        delete tuples_iterator;
        iterator.answer_iterator = NULL;
        if (Cor_exEnv->explain_strm
                && (Cor_exEnv->GlobalRelationOptions & REL_EXPLAIN)){
            // Send a special breakpoint (and leave the pipe open)
            *(Cor_exEnv->explain_strm) << "explain_break(2,0).\n" << flush;
        }
    }
}


// *************************************************************************
void Cor_DerivedRelation::print_name(ostream *out)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_DerivedMethod *method;
    char buf[500];

    for (method = methods; method; method = method->next) {
        if (!method->valid_method)
            continue;

        sprintf(buf, "%s", name.symbol());
        sprint_adornment(buf+strlen(buf), &(method->adornment));

        *out << buf << "/" << arity() << "    : (derived) ";
        *out << ": indexes ";

        Cor_PredAnnotations *cur = NULL;

        if (method->md)
            cur = method->md->index_annotations;

        for (; cur; cur=cur->next()) {
            if ((cur->pred.equals(Cor_CArg(buf)))
                        || (cur->pred.equals(name))) {
                if (!cur->arglist2)
                    cur->arglist1->printon(out);
                else {
                    int mx1 = cur->arglist1->max_vars();
                    int mx2 = cur->arglist1->max_vars();
                    int num_vars = (mx1 > mx2) ? mx1 : mx2;
                    char *argindex = Cor_simplify_index(cur->arglist1,
                                    num_vars, cur->arglist2);
                    if (argindex) 
                        *out << argindex;
                    else {
                        *out << "(";
                        cur->arglist1->printon(out);
                        *out << ")(";
                        cur->arglist2->printon(out);
                        *out << ")";
                    }
                }
                *out << " ";
            }
        }
        *out << endl;
    }
}

// *************************************************************************
void Cor_DerivedRelation::print_facts(ostream *out, Cor_ArgList *) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_DerivedMethod *meth = methods;
    while (meth) {
        if (meth->valid_method) {
            *out << "module " << name.symbol() << ".\n";
            *out << "export " << name.symbol() << "[";
            Cor_print_adornment_brack(meth->adornment, out);
            *out << "].\n";
            Cor_print_Annotations(*(meth->md->module_info), out,
                meth->md->index_annotations,
                meth->md->agg_sel_annotations,
                NULL,
                meth->md->multiset_annotations,
                meth->md->diskrel_annotations,
                meth->md->asyncrel_annotations);
            meth->md->ruleArray->printon(out);
            *out << "end_module.\n";
        }
        meth = meth->next;
    }
}

// *************************************************************************
int Cor_DerivedRelation::insert_new(Cor_ArgList& , Cor_BindEnv*)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    CORAL_error(COR_BAD_TUPLE_INSERT, name.symbol(), "insert_new");
    return(-1);
}

// *************************************************************************
int Cor_orOptionToRelation(Cor_Relation *rel, char *adorn_form,
            RelationOptions Cor_opt)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if ((rel->r_kind != COR_R_DERIVED) && (rel->r_kind != COR_R_ANSWER)) {
        *(Cor_exEnv->err_strm) << 
            "CORAL::Error: Relation is not a derived relation\n";
        return 0;
    }

    Cor_BitVector *bv = new Cor_BitVector(rel->arity());

    int i;    
    for (i = 0; i < rel->arity(); i++, adorn_form++) {
        if (*adorn_form == 0) {
            *(Cor_exEnv->err_strm) << "CORAL::Error: adornment too short.";
            *(Cor_exEnv->err_strm)
                    << "    --- arity of " << rel->name.symbol()
                    << "=" << rel->arity() << ", adorn len="
                    << i << ")\n";
            delete bv;
            return 0;
        }
        if (*adorn_form == 'b' || *adorn_form == '1')
            bv->set(i);
    }

    if (*adorn_form != 0) {
        *(Cor_exEnv->err_strm) << "CORAL::Error: adornment too long.";
        *(Cor_exEnv->err_strm)
                << "    --- arity of " << rel->name.symbol()
                << "=" << rel->arity() << ", adorn len="
                << i << ")\n";
        delete bv;
        return 0;
    }

    Cor_DerivedMethod *meth = ((Cor_DerivedRelation *)rel)->methods;
    while (meth) {
        if (meth->valid_method && (meth->adornment==*bv)) {
            meth->md->rel_options |= Cor_opt;
            delete bv;
            return 1;
        }
        meth = meth->next;
    }
    delete bv;
    *(Cor_exEnv->err_strm) << 
        "CORAL::Error: cannot find the adorned relation\n";
    return 1;
}

// *************************************************************************
int Cor_andOptionToRelation(Cor_Relation *rel, char *adorn_form,
        RelationOptions Cor_opt)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if ((rel->r_kind != COR_R_DERIVED) && (rel->r_kind != COR_R_ANSWER)) {
        *(Cor_exEnv->err_strm) << 
            "CORAL::Error: Relation is not a derived relation\n";
        return 0;
    }

    Cor_BitVector *bv = new Cor_BitVector(rel->arity());

    int i;    
    for (i = 0; i < rel->arity(); i++, adorn_form++) {
        if(*adorn_form == 0) {
            *(Cor_exEnv->err_strm) << "CORAL::Error: adornment too short.";
            *(Cor_exEnv->err_strm)
                << "    --- arity of " << rel->name.symbol()
                << "=" << rel->arity() << ", adorn len="
                << i << ")\n";
            delete bv;
            return 0;
        }
        if (*adorn_form == 'b' || *adorn_form == '1')
            bv->set(i);
    }

    if (*adorn_form != 0) {
        *(Cor_exEnv->err_strm) << "CORAL::Error: adornment too long.";
        *(Cor_exEnv->err_strm) << "    --- arity of " << rel->name.symbol()
            << "=" << rel->arity() << ", adorn len=" << i << ")\n";
        delete bv;
        return 0;
    }

    Cor_DerivedMethod *meth = ((Cor_DerivedRelation *)rel)->methods;
    while (meth) {
        if (meth->valid_method && (meth->adornment==*bv)) {
            meth->md->rel_options &= Cor_opt;
            delete bv;
            return 1;
        }
        meth = meth->next;
    }
    delete bv;
    *(Cor_exEnv->err_strm) << 
        "CORAL::Error: cannot find the adorned relation\n";
    return 1;
}

// *************************************************************************
