// *************************************************************************
//    LinkedRelation.C
// *************************************************************************

#include "LinkedRelation.h"

#include "Globals.h"
#include "RelationList.h"
#include "Debug.h"

// *************************************************************************
Cor_LinkedRelation::Cor_LinkedRelation(int arity, int delta)
    : Cor_StorageRelation(arity, delta), index_set(this) 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // It is essential to have at least one index - a lot of code depends
    // on this.
    
    Cor_BitVector *bv = new Cor_BitVector(arity);

    for (int i = 0; i < arity; i++) {
        // all arguments bound
        bv->set(i);
    }

    Cor_HashIndex *h1 = new Cor_HashIndex(this, bv);
    index_set.insert_index( h1 );
}

// *************************************************************************
Cor_RMark *Cor_LinkedRelation::getMark()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return new Cor_RMark(tlist,tlist->count());
}

// *************************************************************************
void Cor_LinkedRelation::initialize_iterator( Cor_TupleIterator& iterator )
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#    ifdef DO_TRACE
        if (Cor_exEnv->dbg_get_next) {
            *(Cor_exEnv->trace_strm)
                << "\nLinkedRelation::initialize_iterator: ";
            iterator.relation->name->print(NULL, Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << " ( ";
            iterator.arg_list.print(iterator.bindenv, Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << " )\n";
        }
#    endif
    iterator.stack_mark.get_mark();

    Cor_Term t1;
    int all_free = 1;

    for (int i = 0; i < _arity; i++) {
        if (iterator.arg_list[i]->hash(iterator.bindenv) != VarHashValue) {
            all_free = 0;
            break;
        }
    }

    if (!delta_indexed) {
        if (iterator.start_mark) {
            // don't need to go to the hash table
            iterator.ipos = Scanning_New;
            iterator.bkt  = iterator.start_mark->tlist->list();
            iterator.ppos = iterator.start_mark->tindex;
        } else {
            // need to scan from hash table
            // find the appropriate index to be used
            if (all_free) { 
                iterator.bkt  = list();
                iterator.ppos = 0;
                iterator.index = NULL;
            } else 
                iterator.index = index_set.find_index(iterator);
        }
    } else {
        if (!iterator.in_delta()) {
            // Starting on main relation
            if (iterator.start_mark) {
                // go to the delta relations
                iterator.ipos = Scanning_Init;
                iterator.delta_rel_pos =  (iterator.start_mark)->rel_p;
            }
            else {
                // need to scan from hash table
                // find the appropriate index to be used
                if (all_free) {
                    iterator.bkt  = list();
                    iterator.ppos = 0;
                    iterator.index = NULL;
                } else 
                    iterator.index = index_set.find_index(iterator);
            }
        } else {
            // In a delta relation
            if (all_free) {
                iterator.bkt  = list();
                iterator.ppos = 0;
                iterator.index = NULL;
            } else 
                iterator.index = index_set.find_index(iterator);
        }
    }

#    ifdef DO_TRACE
        if (Cor_exEnv->dbg_get_next && !iterator.index) {
            *(Cor_exEnv->trace_strm) << " : using tuple list !\n";
        }
#    endif
}

// *************************************************************************
#define FAIL_RETURN \
{ \
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    iterator.ipos = Scanning_Done;     \
    iterator.set_no_match();    \
    return NULL;            \
}

Cor_Tuple *Cor_LinkedRelation::get_next_tuple(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#    ifdef DO_PROFILE
        if (TableStackCount > 0) {
            ((Cor_TableStack[TableStackCount-1])->moduledata->
                module_info->Cor_GlobalProfileInfo.getnexts)++;
        }
        else {
            (Cor_GlobalProfileInfo.getnexts)++;
        }
#    endif
    
    Cor_BindEnv *result_env = iterator.bindenv;
    iterator.tuple_env = NULL;
    
    Cor_Tuple *result = NULL;

    if (Cor_exEnv->C_interrupt_raised) FAIL_RETURN;

    // initialize
    if (iterator.ipos == Scanning_Init)
        initialize_iterator(iterator);
    else
        iterator.stack_mark.pop_to();


    if (!iterator.index) {
        // an unbound scan using the linked rel !

        Cor_StackMark stack_mark;
        Cor_Tuple    *tuple;

        while (iterator.ppos < iterator.bkt->count()) {

            if (iterator.end_mark
                    && (iterator.end_mark->tindex <= iterator.ppos)) { 
                tuple = NULL;
                break;
            } 

            tuple = iterator.bkt->tuples[iterator.ppos];

            (iterator.ppos)++;

            if (!tuple || tuple->is_deleted())
                continue;

            if (Cor_unify_literal(iterator, tuple, 0))
                return(tuple);
            
            stack_mark.pop_to();
        }

        if (tuple == NULL) { 
            iterator.stack_mark.pop_to();
            FAIL_RETURN;
        }
    } else {
          // use the chosen hash-index : ASSUME DELTA_INDEXING

        if (!iterator.in_delta()) {
            // Scanning from main hash table
            result = iterator.index->get_next_tuple(iterator);
            if (result != NULL)
                return result;
            if (!delta_relations)
                FAIL_RETURN;
            
            // now go into the delta relations
            // Scanning_Init for delta reln
            iterator.ipos = Scanning_Init;
            iterator.delta_rel_pos = delta_relations->reln_chain;
        }

        // delta_rel_pos has to have been set by this point.
        if (delta_relations != NULL) { 
            // iterator scanning deltas, this == main relation
            for(;;) {
                // Loop till successful delta is found
                if (iterator.ipos == Scanning_Init) {
                    if (iterator.delta_rel_pos == NULL) {
                        FAIL_RETURN;
                    }
                    if (iterator.end_mark && (iterator.delta_rel_pos ==
                        (iterator.end_mark)->rel_p))
                    {
                        FAIL_RETURN;
                    }
                }
                result = iterator.delta_rel_pos->reln->get_next_tuple(iterator);
                if (result != NULL)
                    return result;

                // Done with current delta relation.  Got to go to next delta.
                iterator.delta_rel_pos = iterator.delta_rel_pos->next;
                // To start on new delta reln.
                iterator.ipos = Scanning_Init;      
            }
        } else {
            // iterator scanning deltas, this == delta relation
            return iterator.index->get_next_tuple( iterator);
        }
    }
}

#undef FAIL_RETURN

// *************************************************************************
void Cor_LinkedRelation::release(Cor_TupleIterator&)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // Do nothing, since TupleIterator does not maintain extra info to
    // be released in the case of LinkedRelations.
}

// *************************************************************************
Cor_BindEnv *Cor_LinkedRelation::get_next(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Tuple *result = get_next_tuple(iterator);
    if (!result)
        return NULL;
    return iterator.bindenv;
}

// *************************************************************************
int Cor_LinkedRelation::is_subsumed(Cor_Tuple *newtuple,
        Cor_RMark *start_mark, Cor_RMark *end_mark)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_TupleStack *bkt;

    if (start_mark) {
        bkt = start_mark->tlist;
        idx = start_mark->tindex;
    } else {
        bkt = tlist;
        idx = 0;
    } 

    Cor_StackMark stack_mark;

    Cor_Tuple *tuple;

    for (; idx < bkt->count(); idx++) {

        if (end_mark
                && (end_mark->tindex < idx)) { 
            break;

        tuple = bkt->tuples[idx];

        if (!tuple || tuple->is_deleted())
            continue;

        if (Cor_subsumes_tuple(tuple, newtuple, (DISPLAY_INSERTIONS)))
            return(1);

        stack_mark.pop_to();
    }

    return is_agg_subsumed(newtuple, start_mark, end_mark);
}

// *************************************************************************
int Cor_LinkedRelation::insert_raw(Cor_Tuple *tuple)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    //    tuple has been logically deleted
    if (tuple->is_deleted())
        return(0);

    tlist_insert_tuple(tuple);

    // inserts tuple into each index constructed for the relation.
      
    for (int i = 0; i < index_set.num_indices(); i++) {
        index_set[i]->insert_tuple(tuple);
    }
    return(1);
}

// *************************************************************************
void Cor_LinkedRelation::empty_relation(int deleteTuples)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // Empty out relation, but do NOT delete tuples (which may be shared)

    Cor_TupleList::free_all(deleteTuples);
}

// *************************************************************************
void Cor_LinkedRelation::print_name( ostream *out)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *out << name.symbol() << "/" << arity() "  : (base:list)    ";
    index_set.print_adornments(out);
    *out << ": count " << count << endl;
}

// *************************************************************************
