// *************************************************************************
//    StorageRelation.h
// *************************************************************************

#ifndef _StorageRelation_H_
#define _StorageRelation_H_

#include "Relation.h"
#include "TupleStack.h"
#include "TupleList.h"
#include "RMark.h"
#include "ArgList.h"
#include "BitVector.h"
#include "PredAnnotations.h"
#include "AggSelInfo.h"
#include "Tuple.h"
#include "BindEnv.h"
#include "Debug.h"

// *************************************************************************

class Cor_Relation;
class Cor_RelationList;
class Cor_GenericIndex;
class Cor_GenericIndexSet;
class Cor_Context;
class Cor_GoalNode;
class Cor_TupleIterator;
class Cor_QFModuleData;

// *************************************************************************

class Cor_StorageRelation : public Cor_Relation {

  public:

    Cor_TupleStack       tlist;     // A list of tuples

    // auxiliary relation used only with magic relations, for generating
    // goal identifiers if required.
    Cor_StorageRelation *for_supp_index;

      // relation delta fields
    Cor_RMark headMark;
    Cor_RMark *free_marks;       // List of free rmarks.
                                 // Used to reduce calls to new

    Cor_RelationList *delta_relations;
    // above two fields used for determining if the delta relations
    // will be a chain of tuples or a chain of indexed relations

    // List of free relations that can be reused as delta relations.
    Cor_RelationLink *free_reln_list;

    inline int there_are_marks() {  return headMark.next != &headMark; }

    // index and semi-naive related methods
    virtual Cor_GenericIndexSet *get_index_set() {
        return NULL;
    }

    virtual Cor_GenericIndex *create_index_object(Cor_BitVector *) {
        return NULL;
    }

    virtual Cor_GenericIndex *create_index_object(Cor_ArgList *, int,
            Cor_ArgList *) {
        return NULL;
    }

    int add_index(char *s);
    int add_index(Cor_BitVector *s);
    virtual int add_index(Cor_ArgList *args);
    virtual int add_index(Cor_ArgList *pat, int n_var, Cor_ArgList *v_nm);

    int add_all_indices(Cor_StorageRelation *reln);
    int add_all_tuples(Cor_StorageRelation *_reln, int marks_into_account = 0,
        int check_dupl = 0);
    virtual int add_all_goals (Cor_StorageRelation *reln, Cor_Context *context,
        Cor_StorageRelation *owner_rel);

    virtual void add_agg_sel_anno(Cor_PredAnnotations *anno);
    virtual void add_agg_sel_info(Cor_AggSelInfo *newinfo);

    // insert and related methods

    // this is called to actually perform a physical insert of a tuple
    virtual int insert_raw(Cor_Tuple *tuple);

    // insert_new is called to insert a _new_ tuple
    // it therefore has to perform subsumption checks, etc
    // insert if not subsumed.
    virtual int insert_new(Cor_ArgList& args, Cor_BindEnv *env);

    // this checks to see if there are deltas
    virtual int insert_tuple(Cor_Tuple *tuple); 
    int insert(Cor_Tuple *tuple);        // this is probably redundant

                                         // this one is used by ordered search
    int insert_in_context(Cor_GoalNode *g_node, Cor_GoalNode *parent_goal);
    int insert_grouping(Cor_ArgList&args, Cor_BindEnv *env);

    virtual Cor_RMark *getMark();
    virtual void freeMark(Cor_RMark *);
    virtual void free_allMarks();

    virtual int is_subsumed(Cor_Tuple *tuple, Cor_RMark *startmark=NULL,
        Cor_RMark *endmark=NULL);
    virtual int is_agg_subsumed(Cor_Tuple *tuple, Cor_RMark *startmark=NULL,
        Cor_RMark *endmark=NULL);

    virtual int per_iteration_aggregation(Cor_RMark *startmark=NULL,
            Cor_RMark *endmark=NULL);

    virtual void free_all();
    virtual void free_indices();
    virtual int check_subsumption()    { return check_subsum; }
    void set_subsumption(int subsum_flag) {
        check_subsum = subsum_flag;
    }

    virtual Cor_BindEnv *get_next(Cor_TupleIterator&) = 0;
    virtual Cor_Tuple *get_next_tuple(Cor_TupleIterator&) = 0;

    Cor_StorageRelation( int arity, int delta_indexed = -1 );
    Cor_StorageRelation();

    virtual ~Cor_StorageRelation();
};

// *************************************************************************
// A Tuple iterator is used to loop through (part of a) Relation
// Only return those tuples "between" start_mark and end_mark,
// and only those which match 

typedef Cor_StorageRelation * (*InterpDerivedSolver)(Cor_QFModuleData *md,
    Cor_StorageRelation *magic, int old_magic_relation,
    Cor_TupleIterator *iterator=NULL);

Cor_StorageRelation *Cor_AllocateRelation(Cor_CArg& name, int arity,
    int delta_indexed);

// *************************************************************************

#endif _StorageRelation_H_
