#ifdef URL_REL

// *************************************************************************
//    UrlRelation.C
// *************************************************************************

#include <assert.h>
#include <strstream.h>

#include "UrlRelation.h"

#include "Util.h"
#include "Globals.h"
#include "CoralError.h"
#include "Unify.h"
#include "Debug.h"
#include "TupleIterator.h"
#include "ClientProtocol.h"

// *************************************************************************
Cor_UrlRelation::Cor_UrlRelation(char *ln, int arity, char * url, char *rn)
    : Cor_StorageRelation(arity) 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    strcpy(_local_name,ln);
    strcpy(_url, url);
    strcpy(_relation_name, rn); 
    _arity = arity;
    
    // turn off the checking for subsumption. it's done on remote server
    check_subsum = 0;

    if (strncmp(url,"http://",7)) {
        *(Cor_exEnv->err_strm) << "Incorrect url format.\n";
        valid = 0;
    }
    else
        valid = 1;

}

// *************************************************************************
Cor_UrlRelation::~Cor_UrlRelation()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
}

//*************************************************************************
void Cor_UrlRelation::Init(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (iterator.ipos == Scanning_Init) {
        char query[1000];
        ostrstream ost(query, 1000);
        ost << _url << "?assign(answer_style,3).clear(verbose_answer).clear(interactive_mode).";
        ost << "?" << _relation_name << "(";
        for (int i = 0; i < iterator.arg_list.arity(); i++)  {
            Cor_Term term(iterator.arg_list[i], iterator.bindenv);
            Cor_deep_dereference_term(term);
            ost << term.expr;
            if (i == iterator.arg_list.arity() - 1)
                ost << ')';
            else
                ost << ',';
        }
        ost << '.' << '\0';

        int fd, len;
  
        if (strncmp(query,"http://",7)) {
            fprintf(stderr,"Invalid URL\n");
            exit(1);
        }
        fd = open_http(query+5,&len);
        if (fd < 0) {
            fprintf(stderr,"Could not open URL\n");
            exit(1);
        }
        len = read(fd,result,10000); // ?????
        ist = new istrstream(result, 10000);
        
        iterator.stack_mark.get_mark();
        iterator.ipos = Scanning_Known;

    }
    else {
        iterator.stack_mark.pop_to();

    }

}

//*************************************************************************
Cor_Tuple * Cor_UrlRelation::Done(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    iterator.ipos = Scanning_Done;
    iterator.set_no_match();

    return NULL;
}

//*************************************************************************
Cor_Tuple * Cor_UrlRelation::GetNext()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char Buf[1000];
    while ((ist->getline(Buf,1000))) {
        if (!Buf) return NULL;
        if (!strncmp(_relation_name, Buf, strlen(_relation_name))) {
            Cor_CArg relname(_relation_name);
            int ar = _arity;
            Cor_Tuple * tuple = Cor_parseFact(Buf,_relation_name, &ar);
            if (tuple) 
                return tuple;
        }
    }
    return NULL;
}

//*************************************************************************
Cor_Tuple *Cor_UrlRelation::get_next_tuple(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (Cor_exEnv->C_interrupt_raised)
      return Done(iterator);

    Init(iterator);

    Cor_Tuple * tuple = GetNext();

    if (tuple && Cor_unify_literal(iterator, tuple, Cor_exEnv->dbg_get_next)) {
        iterator.reset_no_match();
        return tuple;
    }
    else 
        return Done(iterator);
    
}

// *************************************************************************
Cor_BindEnv *Cor_UrlRelation::get_next(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Tuple *result = get_next_tuple(iterator);
    if (!result) return NULL;
    return iterator.bindenv;
}

// *************************************************************************
int Cor_UrlRelation::insert_tuple(Cor_Tuple *tuple)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char query[1000];
    ostrstream ost(query,1000);
    ost << _url << "?insert(" << _relation_name << "(" << *tuple << "))." << '\0';
    http_url(query);

    return 0; // let the caller decide whether to delete the tuple
}

// *************************************************************************
void Cor_UrlRelation::print_name(ostream *out)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *out << "local_name = " << _local_name << endl;
    *out << "url = " << _url << endl;
}

// *************************************************************************
void Cor_UrlRelation::release(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Done(iterator);
}

// *************************************************************************
void Cor_UrlRelation::empty_relation(int deleteTuples)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!deleteTuples) return;
    
    // ??? it's really a relation on the remote server. we should not empty it.
}

// *************************************************************************
int Cor_UrlRelation::tuple_delete(Cor_Tuple* tuple)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char query[1000];
    ostrstream ost(query, 1000);
    ost << _url << "?delete(" << _relation_name << "(" << *tuple << ")).";
    http_url(query);
    
    return 1; 
}

// *************************************************************************

/*
  This file contributed by the Condor project and modified locally.
  See attached copyright notice.
*/

/*
  $Id: UrlRelation.C,v 1.2 1996/09/09 01:01:01 flisakow Exp $

  $Log: UrlRelation.C,v $
  Revision 1.2  1996/09/09 01:01:01  flisakow
      Added new CallStk class (and friends) to implement a call-stack
      and call-trace mechanism for debugging.

  Revision 1.1  1996/05/10 20:44:02  whuang
  added asyn_sock_rel url_rel and wodi_client for REEF

  Revision 1.1  1996/01/13 02:25:54  jussi
  Initial revision.
*/

/* 
** Copyright 1995 by Miron Livny and Jim Pruyne
** 
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted,
** provided that the above copyright notice appear in all copies and that
** both that copyright notice and this permission notice appear in
** supporting documentation, and that the names of the University of
** Wisconsin and the copyright holders not be used in advertising or
** publicity pertaining to distribution of the software without specific,
** written prior permission.  The University of Wisconsin and the 
** copyright holders make no representations about the suitability of this
** software for any purpose.  It is provided "as is" without express
** or implied warranty.
** 
** THE UNIVERSITY OF WISCONSIN AND THE COPYRIGHT HOLDERS DISCLAIM ALL
** WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
** OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE UNIVERSITY OF
** WISCONSIN OR THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT
** OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
** OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
** OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
** OR PERFORMANCE OF THIS SOFTWARE.
** 
** Author:  Jim Pruyne
**
*/ 

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include "machdep.h"

extern int errno;

#define HTTP_PORT	80

#define HTTP_GET_FORMAT "GET %s HTTP/1.0"
#define INT_CR 13
#define INT_LF 10

#define HTTP_C_LENGTH "content-length"
#define HTTP_C_LENGTH_LEN strlen(HTTP_C_LENGTH)

#define HTTP_VERSION "HTTP/1.0"
#define HTTP_VERSION_LEN strlen(HTTP_VERSION)

#define BUF_LENGTH 1024

static void
tolowcase(char *s)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
  int c,i=0,delta;
  while ((c = s[i]) != '\0') {
    if (((delta = (c - 'A')) >= 0) && (delta < 26)) s[i] = 'a' + delta;
    i++;
    if (i == BUF_LENGTH) break;
  }
}

/* read lines terminated by CRLF, replacing CR, LF by \0 */
static int
readline(int fd, char *buf)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
	int		count = 0;
	int		rval;

	for(;;) {
		if ((rval = read(fd, buf, 1)) < 0) {
			return rval;
		}
		else if (rval == 0) {
		  /* shouldn't really get here */
		  *buf = '\0';
		  break;
		}

		if (*buf != INT_LF) {
			if (*buf == INT_CR) *buf = '\0';
			else {
		  	  count++;
			  if (count == BUF_LENGTH) {
			    /* hack */
			    *buf = '\0';
			    count--;
			    break;
			  }  
		  	  buf++;
			}
		} else {
			*buf = '\0';
			break;
		}
	}
	return	count;
}


extern
int open_http( const char *name, int * bytes_in_body)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

	struct sockaddr_in	sin;
	int		sock_fd;
	int		status;
	char	*port_sep;
	char	*end_of_addr;
	char    *clength_start;
	char    *stat_code_start, *reason_phrase_start;
	int stat_code, l;
	int		port_num = HTTP_PORT;
	struct hostent *he;
	char	http_cmd[BUF_LENGTH];
	

	sock_fd = socket(AF_INET, SOCK_STREAM, 0);
	if (sock_fd < 0) {
		fprintf(stderr, "socket() failed, errno = %d\n", errno);
		fflush(stderr);
		return sock_fd;
	}

	if (name[0] != '/' || name[1] != '/') {
		return -1;
	}

	/* Skip those two leading '/'s */
	name++;
	name++;

	port_sep = strchr(name, ':');
	end_of_addr = strchr(name, '/');
	if (end_of_addr == 0) {
		return -1;
	}
	*end_of_addr = '\0';

	if (port_sep) {
		*port_sep = '\0';
		port_sep++;
		port_num = atoi(port_sep);
	}
	sin.sin_port = htons(port_num);

	he = gethostbyname( name );
	if ( he ) {
		sin.sin_family = he->h_addrtype;
		sin.sin_addr = *((struct in_addr *) he->h_addr_list[0]);
	} else {
		return -1;
	}

	*end_of_addr = '/';

	status = connect(sock_fd, (struct sockaddr *) &sin, sizeof(sin));
	if (status < 0) {
		fprintf(stderr, "http connect() FAILED, errno = %d\n", errno);
		fflush(stderr);
		return status;
	}

	name = end_of_addr;
	sprintf(http_cmd, HTTP_GET_FORMAT, name);
	l = strlen(http_cmd);
	http_cmd[l++] = INT_CR;
	http_cmd[l++] = INT_LF;
	http_cmd[l++] = INT_CR;
	http_cmd[l++] = INT_LF;
	http_cmd[l] = '\0';
	write(sock_fd, http_cmd, l);

	/* get status line */
	if (readline(sock_fd, http_cmd) <= 0) {
	  fprintf(stderr,"Bad http connection\n");
	  close(sock_fd);
	  return(-1);
	}

	if (strncmp(http_cmd, HTTP_VERSION, HTTP_VERSION_LEN)) {
	  fprintf(stderr,"Bad http header received\n");
	  close(sock_fd);
	  return(-1);
	}

	if (!(stat_code_start = strchr(http_cmd,' '))) { 
	  fprintf(stderr,"Bad http header received\n");
	  close(sock_fd);
	  return(-1);
	}
	stat_code_start++;

	if (!(reason_phrase_start = strchr(stat_code_start,' '))) { 
	  fprintf(stderr,"Bad http header received\n");
	  close(sock_fd);
	  return(-1);
	}
	*reason_phrase_start = '\0';
	reason_phrase_start++;

	stat_code = atoi(stat_code_start);
	if (stat_code < 200) {
	  fprintf(stderr,"Unused status code (%d) received\n", stat_code);
	  close(sock_fd);
	  return(-1);
	}
	else if (stat_code < 300) {
	  /* cool.. */
	}
	else if (stat_code < 400) {
	  fprintf(stderr,"Redirection error: %s\n", reason_phrase_start);
	  close(sock_fd);
	  return(-1);
	}
	else if (stat_code < 500) {
	  fprintf(stderr,"Client error: %s\n", reason_phrase_start);
	  close(sock_fd);
	  return(-1);
	}
	else if (stat_code < 600) {
	  fprintf(stderr,"Server error: %s\n", reason_phrase_start);
	  close(sock_fd);
	  return(-1);
	}
	else {
	  fprintf(stderr,"Strange status code (%d) received\n", stat_code);
	  close(sock_fd);
	  return(-1);
	}


	/* Skip the header part, but extract Content-Length */
	*bytes_in_body = -1;
	while (readline(sock_fd, http_cmd) > 0) {
	  tolowcase(http_cmd);
	  if (!strncmp(http_cmd, HTTP_C_LENGTH, HTTP_C_LENGTH_LEN)) {
	    if ((clength_start = strchr(http_cmd,':'))) {
	      clength_start++;
	      while (*clength_start == ' ') clength_start++;
	      *bytes_in_body = atoi(clength_start);
	    }
	  }  
	}
	return sock_fd;
}

void http_url(char * url)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
  char url_data[BUF_LENGTH];
  int fd, len;
  
  if (strncmp(url,"http://",7)) {
    fprintf(stderr,"Invalid URL\n");
    exit(1);
  }
  fd = open_http(url+5,&len);
  if (fd < 0) {
    fprintf(stderr,"Could not open URL\n");
    exit(1);
  }
  while ((len = read(fd,url_data,BUF_LENGTH)) > 0)
	  write(1,url_data,len);
  close(fd);
}

#endif // URL_REL
