// *************************************************************************
//    PredAnnotations.C
// *************************************************************************

#include "PredAnnotations.h"

#include "Unparse.h"
#include "Globals.h"
#include "Debug.h"

// *************************************************************************
Cor_PredAnnotations::Cor_PredAnnotations(Cor_CArg& pred1,
        Cor_PredAnnotations *next1)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    pred  = pred1;
    _next = next1;
    op_arg.init();
    arglist1 = arglist2 = NULL;
    adorn    = NULL;
    agg_type = AggSel_Regular;
}

// *************************************************************************
Cor_PredAnnotations::~Cor_PredAnnotations()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // Causes trouble for magic - SPF
    // delete arglist1;
    // delete arglist2;

    delete adorn;
}

// *************************************************************************
//  To conviently delete a list of these things
//  Note: This routine _must_ handle the case where (this == NULL)
void Cor_PredAnnotations::delete_list()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_PredAnnotations *curr, *prev = NULL;
    
    for (curr = this; curr; curr = curr->_next) {
        delete prev;
        prev = curr;
    }

    delete prev;
}

// *************************************************************************
//  if "adorned" is not set, the field "adorn" is printed in
//  parentheses instead of with an "_"
// *************************************************************************

void Cor_PredAnnotations::print(ostream *file, int adorned)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (adorned) {
        Cor_fprint_pred_name(file, pred.symbol(), adorn);
    } else {
        Cor_fprint_pred_name(file, pred.symbol(),
                (Cor_BitVector *) NULL);
        if (adorn) {
            *file << "[";
            for (int i = 0; i < adorn->len(); i++)
                *file << (adorn->test(i) ? "b" : "f");
            *file << "]";
        }
    }

    if (arglist1) {
        *file << "(";
        arglist1->print(NULL,file);
        *file << ") ";
    }

    if (arglist2) {
        *file << "(";
        arglist2->print(NULL,file);
        *file << ") ";
    }

    if (!op_arg.is_empty()) {
        op_arg.print(NULL,file);
    }
}

// *************************************************************************
