// *************************************************************************
//    Solvers.C
// *************************************************************************

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <ctype.h>

#include "Solvers.h"

#include "Config.h"
#include "Globals.h"
#include "CoralError.h"
#include "Interface.h"

#include "ArgList.h"
#include "BuiltinSymbols.h"
#include "TupleIterator.h"
#include "Tuple.h"
#include "Term.h"
#include "BindEnv.h"
#include "HashSimpleRelation.h"
#include "Execute.h"
#include "DerivedRelation.h"
#include "BuiltinUtil.h"
#include "Util.h"
#include "Database.h"
#include "Suspend.h"
#include "Suspension.h"
#include "SockStream.h"
#include "SortedRelation.h"
#include "ClientProtocol.h"

#include "Debug.h"

// *************************************************************************

#ifdef    SAVE_INTER
static int unlink(char *path)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    static cnt = 0;
    char cmmd[100];
    *cerr_strm << "Not unlinking " <<  path << " (moving to corsave." <<  cnt << ")\n";
    sprintf(cmmd, "/bin/mv %s /tmp/corsave.%d\n", path, cnt++);
    system(cmmd);
    return (0);
}

#endif                                // SAVE_INTER

// *************************************************************************
#define FAIL_RETURN {iterator.set_no_match(); return(iterator.bindenv);}

Cor_BindEnv *Cor_InsertFactSolver(Cor_BuiltinRelation &,
          Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef DO_TRACE
    if (Cor_exEnv->dbg_get_next) {
        *(Cor_exEnv->trace_strm) << "Entering %s\n", "insert";
    }
#endif

    if (iterator.arg_list.count() > 0) {

        FOR_EACH_ARG(arg, iterator.arg_list) {
            Cor_Term term1(*arg, iterator.bindenv);

            Cor_deep_dereference_term(term1);

            if (!term1.expr.is_functor()) {
                CORAL_error(COR_BAD_ARGLIST, "Arguments must be functors",
                            "InsertFactSolver");
                term1.printon(Cor_exEnv->err_strm);
                *(Cor_exEnv->err_strm) << endl;
                FAIL_RETURN;
            }

            Cor_CArg tmp = Cor_CArg(term1.expr.functor());
            Cor_Relation *rel =
                Cor_find_local_or_external_relation(tmp, term1.expr.arity());

            // term1 should be ground now, bindenv is not needed? -SPF
            rel->insert_new(term1.expr.args(), iterator.bindenv);
        } END_EACH_ARG
    }

    return (iterator.bindenv);
}

#undef FAIL_RETURN

// *************************************************************************
#define FAIL_RETURN {iterator.set_no_match(); return(iterator.bindenv);}

Cor_BindEnv *Cor_DeleteFactSolver(Cor_BuiltinRelation &,
        Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef DO_TRACE
    if (Cor_exEnv->dbg_get_next) {
        *(Cor_exEnv->trace_strm) << "Entering %s\n", "delete";
    }
#endif

    if (iterator.arg_list.count() > 0) {

        FOR_EACH_ARG(arg, iterator.arg_list) {

            Cor_Term term1(*arg, iterator.bindenv);
            Cor_deep_dereference_term(term1);

            if (!term1.expr.is_functor()) {
                CORAL_error(COR_BAD_ARGLIST, "Arguments must be functors",
                            "DeleteFactSolver");
                term1.printon(Cor_exEnv->err_strm);
                *(Cor_exEnv->err_strm) << endl;
                FAIL_RETURN;
            }

            if (term1.expr.is_functor() && term1.expr.functor())  {
                Cor_CArg tmp = Cor_CArg(term1.expr.functor());
                Cor_Relation *rel =
                        Cor_find_local_or_external_relation(tmp,
                        term1.expr.arity());

                Cor_ArgList del_al = term1.expr.args();
                Cor_Tuple *tuple1 = new Cor_Tuple(del_al);

                tuple1->instantiate(iterator.bindenv);
                rel->tuple_delete(tuple1);
                delete tuple1;
            } else {
                FAIL_RETURN;
            }

        }
        END_EACH_ARG
    }
    return (iterator.bindenv);
}

#undef FAIL_RETURN

// *************************************************************************
#define FAIL_RETURN {iterator.set_no_match(); return(iterator.bindenv);}

Cor_BindEnv *Cor_addIndexSolver1(Cor_BuiltinRelation &,
        Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // Arglist should have one arg - a funcarg with relation name, and
    // single arg with the binding pattern.

#ifdef DO_TRACE
    if (Cor_exEnv->dbg_indexing) {
        *(Cor_exEnv->trace_strm) << "Entering "
            << Cor_MakeIndexSymbol->symbol() << endl;
    }
#endif

    if (iterator.arg_list.arity() != 1) {
        *(Cor_exEnv->err_strm)
            << Cor_MakeIndexSymbol->symbol()
            << ": Number of arguments must be 1.\n";
        FAIL_RETURN;
    }
    Cor_Term term1(iterator.arg_list[0], iterator.bindenv);
    Cor_deep_dereference_term(term1);
    if (!term1.expr.is_functor()) {
        *(Cor_exEnv->err_strm)
            << "CORAL :: Error -- bad argument to "
            << Cor_MakeIndexSymbol->symbol() << " =" << term1 << endl;
        FAIL_RETURN;
    }
    Cor_CArg *fa = &(term1.expr);
    if ((fa->arity() != 1) || !fa->args()[0].is_symbol()) {
        *(Cor_exEnv->err_strm)
            << "CORAL :: Error -- bad argument to "
            << Cor_MakeIndexSymbol->symbol() << " =" << term1 << endl;
        FAIL_RETURN;
    }
    char *adorn = fa->args()[0].symbol();

    Cor_Relation *indrel = find_relation(fa->functor(), strlen(adorn));
    if (!indrel) {
        *(Cor_exEnv->err_strm)
            << "CORAL :: Error - "
            << Cor_MakeIndexSymbol->symbol()
            << ": relation " << fa->functor()
            << "(arity " << strlen(adorn) << ") not found.\n";
        FAIL_RETURN;
    }
    if (indrel->r_kind != COR_R_ANSWER) {
        *(Cor_exEnv->err_strm)
            << "CORAL :: Error - "
            << Cor_MakeIndexSymbol->symbol()
            << ": relation " << fa->functor()
            << "(arity " << strlen(adorn) << ") not base.\n";
        FAIL_RETURN;
    }
    if (((Cor_StorageRelation *) (indrel))->add_index(adorn) == COR_I_SUCCESS)
        iterator.reset_no_match();
    else
        iterator.set_no_match();

    return (iterator.bindenv);
}

#undef FAIL_RETURN

// *************************************************************************
#define FAIL_RETURN {iterator.set_no_match(); return(iterator.bindenv);}

Cor_BindEnv *Cor_addIndexSolver2(Cor_BuiltinRelation &,
        Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    //    Arglist should have two args - a funcarg with relation name (with a
    //    single arg with the binding pattern) and a list of bound vars.

#ifdef DO_TRACE
    if (Cor_exEnv->dbg_indexing) {
        *(Cor_exEnv->trace_strm) <<
                          "Entering " <<  Cor_MakeIndexSymbol->symbol() << "\n";
    }
#endif

    if (iterator.arg_list.arity() != 2) {
        *(Cor_exEnv->err_strm) << Cor_MakeIndexSymbol->symbol()
            << ": Number of arguments must be 1 or 2.\n";
        FAIL_RETURN;
    }
    Cor_Term term1(iterator.arg_list[0], iterator.bindenv);
    Cor_deep_dereference_term(term1);
    if (!term1.expr.is_functor()) {
        *(Cor_exEnv->err_strm) << "CORAL :: Error -- bad argument to "
             << Cor_MakeIndexSymbol->symbol() << " =";
        term1.printon(Cor_exEnv->err_strm);
        *(Cor_exEnv->err_strm) << endl;
        FAIL_RETURN;
    }
    Cor_CArg *fa = &(term1.expr);

    Cor_Term term2(iterator.arg_list[1], iterator.bindenv);
    Cor_deep_dereference_term(term2);

    Cor_Relation *indrel = find_relation(fa->functor(), fa->arity());
    if (!indrel) {
        *(Cor_exEnv->err_strm)
            << "CORAL :: Error - " << Cor_MakeIndexSymbol->symbol()
            << ": relation " << fa->functor()
            << "(arity " << fa->arity() << ") not found\n";
        FAIL_RETURN;
    }
    *(Cor_exEnv->err_strm) <<
   "CORAL:: add_index - pattern form index addition not fully supported yet\n";
    /**********
    if (indrel->add_index(term1.expr->argsterm2.expr) == COR_I_SUCCESS)
        iterator.reset_no_match();
    else iterator.set_no_match();
    **********/

    return (iterator.bindenv);
}

#undef FAIL_RETURN

// *************************************************************************
Cor_BindEnv *Cor_CutSolver(Cor_BuiltinRelation &, Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (Cor_CurrModuleInfo->UsePipelining && iterator.calling_pei)
        iterator.calling_pei->cut_crossed = 1;
    return (iterator.bindenv);
}

// *************************************************************************
#define FAIL_RETURN {iterator.set_no_match(); return(iterator.bindenv);}

Cor_BindEnv *Cor_CoralRootSolver(Cor_BuiltinRelation&,
                Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (iterator.arg_list.count() == 0) {
        *(Cor_exEnv->out_strm) << "CoralRoot is: "
                << Cor_exEnv->coral_path << endl;
        return(iterator.bindenv);
    }

    if (iterator.arg_list.count() != 1) {
        CORAL_error(COR_BAD_ARITY, "One argument to coralroot needed", NULL);
        FAIL_RETURN;
    } else {
        Cor_Term term1 (iterator.arg_list[0], iterator.bindenv);
        Cor_deep_dereference_term (term1);

        if (term1.expr.is_var()) {
            Cor_CArg crarg = Cor_CArg(Cor_exEnv->coral_path);
            Cor_Term  crterm = Cor_Term(&crarg);
            iterator.bindenv->bind(term1.expr.var_num(), crterm);
            return(iterator.bindenv);
        }

        if (!term1.expr.is_symbol() ) {
            CORAL_error(COR_BAD_ARG_TYPE, NULL, __FUNCTION__);
            term1.printon(Cor_exEnv->err_strm);
            *(Cor_exEnv->err_strm) << endl;
            FAIL_RETURN;
        }

        int len = strlen( term1.expr.symbol() );
        char *cr = new char [len+1];
        strcpy(cr,  term1.expr.symbol());
        cr[len] = '\0';

        delete [] Cor_exEnv->coral_path;
        Cor_exEnv->coral_path = cr;
    }

    return(iterator.bindenv);
}

#undef FAIL_RETURN
// *************************************************************************
#define FAIL_RETURN {iterator.set_no_match(); return(iterator.bindenv);}

Cor_BindEnv *Cor_AliasSolver(Cor_BuiltinRelation &,
        Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (iterator.arg_list.count() != 2) {
        CORAL_error(COR_BAD_ARITY, "2 arguments to alias needed", NULL);
        FAIL_RETURN;
    }
    Cor_Term term1(iterator.arg_list[0], iterator.bindenv);
    Cor_deep_dereference_term(term1);
    if (!term1.expr.is_symbol()) {
        *(Cor_exEnv->err_strm)
            << "CORAL :: Error -- bad first argument to "
            << Cor_AliasSymbol->symbol() << " =" << term1 << endl;
        FAIL_RETURN;
    }
    Cor_Term term2(iterator.arg_list[1], iterator.bindenv);
    Cor_deep_dereference_term(term2);
    if (!term2.expr.is_symbol()) {
        *(Cor_exEnv->err_strm)
            << "CORAL :: Error -- bad second argument to "
            << Cor_AliasSymbol->symbol() << " =" << term2 << endl;
        FAIL_RETURN;
    }
    Cor_BuiltinRelation *tmp = (Cor_BuiltinRelation *)
    Cor_BuiltinDB->RelationTable->lookup(term2.expr);

    if (!tmp) {
        *(Cor_exEnv->err_strm)
            << "Specified builtin relation " << term2.expr.symbol()
            << " does not exist.\n";
        FAIL_RETURN;
    }

    Cor_BuiltinRelation *dummy = new Cor_BuiltinRelation(tmp->arity(),
                                              &(term1.expr), tmp->solver);

    if (dummy)       // Eliminate "unused" warning
        ;            // Nothing

    return(iterator.bindenv);
}

#undef FAIL_RETURN

// *************************************************************************
Cor_BindEnv *Cor_HelpSolver(Cor_BuiltinRelation &, Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (iterator.arg_list.count() == 0) {
        char name_buffer[1000];
        sprintf(name_buffer, "%s%s/general.help",
                Cor_exEnv->coral_path, HELP_PATH);
        if (Cor_DisplayFile(name_buffer))
            CORAL_error(COR_FILE_NOT_FOUND, "Sorry, no help available.", NULL);
    } else {
        Cor_Term term1(iterator.arg_list[0], iterator.bindenv);
        Cor_deep_dereference_term(term1);
        if (!term1.expr.is_symbol()) {
            CORAL_error(COR_BAD_ARG_TYPE, NULL, "HelpSolver");
            term1.printon(Cor_exEnv->err_strm);
            *(Cor_exEnv->err_strm) << endl;
            iterator.set_no_match();
            return(iterator.bindenv);
        }

        int line_num = 0;
        char *help_name = Cor_FindHelpFile(&(term1.expr), &line_num);

        if (!help_name)
            help_name = term1.expr.symbol();

        char name_buffer[1000];
        sprintf(name_buffer, "%s%s/%s.help", Cor_exEnv->coral_path, HELP_PATH,
                help_name);
        if (Cor_DisplayFile(name_buffer)) {
            *(Cor_exEnv->err_strm)
                << "Sorry, no help available for the topic '"
                << help_name << "'\n";
        }
    }

    return(iterator.bindenv);
}

// *************************************************************************
Cor_BindEnv *Cor_HistorySolver(Cor_BuiltinRelation &,
        Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (iterator.arg_list.count() == 0) {
        (Cor_parseEnv->history)->display(Cor_exEnv->err_strm);
        return (iterator.bindenv);
    }

    if (iterator.arg_list.count() != 1) {
        CORAL_error(COR_BAD_ARITY, "1 argument to history needed", NULL);
        iterator.set_no_match();
        return (iterator.bindenv);
    }
    Cor_Term term1(iterator.arg_list[0], iterator.bindenv);
    Cor_deep_dereference_term(term1);
    if (!term1.expr.is_num()) {
        CORAL_error(COR_BAD_ARG_TYPE, NULL, "HistorySolver");
        term1.printon(Cor_exEnv->err_strm);
        *(Cor_exEnv->err_strm) << endl;
        iterator.set_no_match();
        return (iterator.bindenv);
    }
    char *tmp = (Cor_parseEnv->history)->getCommand(make_int(term1.expr));

    if (!tmp) {
        *(Cor_exEnv->err_strm) << "CORAL::Error: out of history range!\n";
        iterator.set_no_match();
        return (iterator.bindenv);
    }
    Cor_parseStack->handle_new_buff(tmp, "buffer");
    Cor_parseStack->parse(Cor_exEnv->err_strm, 1,
                          "syntax error -- skipping to next period\n");

    // Now modify the history stack
    (Cor_parseEnv->history)->addCommand(tmp, Cor_parseEnv->command_count++);
    Cor_parseEnv->is_history_command = 1;

    return (iterator.bindenv);
}

// *************************************************************************
Cor_BindEnv *Cor_DisplayDefaultsSolver(Cor_BuiltinRelation &,
        Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (iterator.arg_list.count() != 0) {
        CORAL_error(COR_BAD_ARITY, "no arguments to display_defaults needed",
                    NULL);
    }
    char name[100];
    char *tmp_file_name = Cor_mk_tmp_file_name(name);
    CoralStream *tmp_file = Cor_open_outfile(tmp_file_name);
    if (!tmp_file) {
        iterator.set_no_match();
        return (iterator.bindenv);
    }
    Cor_exEnv->display(tmp_file);
    delete tmp_file;
    Cor_DisplayFile(tmp_file_name);

    unlink(tmp_file_name);
    return (iterator.bindenv);
}

// *************************************************************************
Cor_BindEnv *Cor_ResetDefaultsSolver(Cor_BuiltinRelation &,
         Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (iterator.arg_list.count() != 0) {
        CORAL_error(COR_BAD_ARITY,
                    "no arguments to reset_defaults needed", NULL);
    }
    if (coralConfig->server_mode)
        Cor_exEnv->init((CoralStream *) NULL,
            (CoralStream *) NULL, (CoralStream *) NULL);
    else
        Cor_exEnv->init();
    return (iterator.bindenv);
}

// *************************************************************************
Cor_BindEnv *Cor_QuitSolver(Cor_BuiltinRelation &, Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_exEnv->C_interrupt_raised = 2;        // quit().

    Cor_exEnv->quit = 1;

    return (iterator.bindenv);
}

// *************************************************************************

#define FAIL_RETURN  {it.set_no_match(); delete [] terms; return(NULL);}

Cor_BindEnv *Cor_ArgsSolver(Cor_BuiltinRelation&, Cor_TupleIterator& it)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_ArgType_Error err;
    Cor_Term *terms = NULL;
    int   j;

    if (it.arg_list.count() <= 1) {
        *(Cor_exEnv->err_strm)
            << "The 'args' builtin requires at least two arguments.\n";
        FAIL_RETURN;
    }

    terms = new Cor_Term [it.arg_list.count()];

    err = Cor_CheckArgTypes(it.arg_list, 2, -1, "fa", terms, 1, it.bindenv);

    if (err == Coral_Wrong_Type_Args) {
        *(Cor_exEnv->err_strm)
            << "The first argument to args() must be a functor.\n";
        FAIL_RETURN;
    }

    if (err != Coral_Args_Okay) {
        *(Cor_exEnv->err_strm)
            << "Incorrect arguments to the args() builtin.\n";
        FAIL_RETURN;
    }

    int len = terms[0].expr.arity();
    int maxvar = it.bindenv->max_var_number();

    Cor_CArg *fname = terms[0].expr.cfunctor();
    Cor_CArg new_arg = Cor_CArg( *fname, len );
    Cor_CArg newvar;

    for (j=1; j <= len; j++) {

        if ( j <= it.arg_list.count()) {
            (new_arg.args())[j-1] = terms[j].expr;
        } else {
              // Fill in with a variable we don't care about
            newvar.set_var(maxvar++);
            (new_arg.args())[j-1] = newvar;
        }
    }

    Cor_Term new_term = Cor_Term(new_arg,it.bindenv);

    cerr << "The new term is: " << new_term << endl;

    Cor_StackMark stackmark;

    if (!Cor_unify_args(terms[0], new_term)) {
        stackmark.pop_to();
        FAIL_RETURN;
    }

    delete [] terms;
    return(it.bindenv);
}

#undef FAIL_RETURN

// *************************************************************************

#define FAIL_RETURN  {iterator.set_no_match(); return(NULL);}

Cor_BindEnv *Cor_UnivSolver(Cor_BuiltinRelation &, Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (iterator.arg_list.count() != 2) {
        *(Cor_exEnv->err_strm) << "Two arguments required for Univ.\n";
        FAIL_RETURN;
    }
    Cor_Term term1(iterator.arg_list[0], iterator.bindenv);
    Cor_deep_dereference_term(term1);

    Cor_Term term2(iterator.arg_list[1], iterator.bindenv);
    Cor_deep_dereference_term(term2);

    if (term1.expr.is_var()) {
        if (!is_list(term2.expr)) {
            *(Cor_exEnv->err_strm) << "Second argument must be a list\n";
            FAIL_RETURN;
        }
        Cor_Term first_arg_term(*(make_car(term2.expr)), iterator.bindenv);
        Cor_deep_dereference_term(first_arg_term);

        if (!first_arg_term.expr.is_symbol()) {
            *(Cor_exEnv->err_strm) <<
                         "First element of list must be functor name\n";
            FAIL_RETURN;
        }
        Cor_Term rest_arg_term(*(make_cdr(term2.expr)), iterator.bindenv);
        Cor_deep_dereference_term(rest_arg_term);

        if ( rest_arg_term.expr.is_cons() ) {
            Cor_ArgList *arg_list;
            Cor_ArgList *temp_list = new Cor_ArgList(1);
            Cor_Term temp_term;
            temp_term.expr = *(make_car(rest_arg_term.expr));
            temp_term.bindenv = iterator.bindenv;
            Cor_deep_dereference_term(temp_term);
            arg_list = new Cor_ArgList(1);
            (*arg_list)[0] = temp_term.expr;
            int count = 1;
            for (;;) {
                rest_arg_term.expr = *(make_cdr(rest_arg_term.expr));
                rest_arg_term.bindenv = iterator.bindenv;
                Cor_deep_dereference_term(rest_arg_term);

                if (rest_arg_term.expr.is_cons()) {
                    temp_term.expr = *(make_car(rest_arg_term.expr));
                    temp_term.bindenv = iterator.bindenv;
                    Cor_deep_dereference_term(temp_term);
                    (*temp_list)[0] = temp_term.expr;
                    arg_list = concat_arglists(arg_list, temp_list);
                    count++;
                } else if (rest_arg_term.expr.is_nil())
                    break;
                else {
                    *(Cor_exEnv->err_strm) << "Error in list argument.\n";
                    return (NULL);
                }
            }

            Cor_CArg *fun_arg = new Cor_CArg(first_arg_term.expr, count);
            int i;

            for (i = 0; i < count; i++)
                (fun_arg->args())[i] = (*arg_list)[i];

            Cor_Term new_term(fun_arg, iterator.bindenv);
            Cor_StackMark stackmark;
            if (!Cor_unify_args(term1, new_term)) {
                stackmark.pop_to();
                FAIL_RETURN;
            }
            return (iterator.bindenv);
        } else if (rest_arg_term.expr.is_nil()) {
            Cor_CArg *fun_arg = new Cor_CArg(first_arg_term.expr, 0);
            Cor_Term new_term(fun_arg, iterator.bindenv);
            Cor_StackMark stackmark;
            if (!Cor_unify_args(term1, new_term)) {
                stackmark.pop_to();
                FAIL_RETURN;
            }
            return (iterator.bindenv);
        } else {
            *(Cor_exEnv->err_strm) <<
                               "List type needed as second argument.\n";
            FAIL_RETURN;
        }
    }

    if (!term1.expr.is_functor()) {
        *(Cor_exEnv->err_strm) << "First argument must be a functor\n";
        FAIL_RETURN;
    }

    int num = term1.expr.args().arity();
    Cor_CArg *name_arg = make_arg(functor_name(term1.expr));
    if (num == 0) {
        // make list with just the name
        Cor_CArg *arg_list = make_cons(name_arg, NULL);

        Cor_Term new_term(*arg_list, iterator.bindenv);
        Cor_StackMark stackmark;
        if (!Cor_unify_args(term2, new_term)) {
            stackmark.pop_to();
            FAIL_RETURN;
        }
        return (iterator.bindenv);
    }
    // else forming the list

    Cor_CArg *arg_list = make_cons(&(term1.expr.args()[num - 1]), NULL);
    int i;
    for (i = (num - 2); i >= 0; i--)
        arg_list = make_cons(&(term1.expr.args()[i]), arg_list);

    arg_list = make_cons(name_arg, arg_list);

    Cor_Term new_term(*arg_list, iterator.bindenv);
    Cor_StackMark stackmark;
    if (!Cor_unify_args(term2, new_term)) {
        stackmark.pop_to();
        FAIL_RETURN;
    }
    return (iterator.bindenv);
}

#undef FAIL_RETURN

// *************************************************************************
Cor_BindEnv *Cor_CallSolver(Cor_TupleIterator& it)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_BindEnv *env = it.bindenv;
    Cor_BindEnv *newenv;

    Cor_Relation *rel;
    Cor_TupleIterator *iter;

    if (it.ipos++ == 0) {
        it.stack_mark.get_mark();
        int num = it.arg_list.arity();
        if (num != 1) {
            *(Cor_exEnv->err_strm) << "Call takes only one argument.\n";
            it.set_no_match();
            return(NULL);
        }

        Cor_Term term(it.arg_list[0], env);
        Cor_deep_dereference_term(term);

        if (!term.expr.is_functor()) {
            *(Cor_exEnv->err_strm)
                    <<"Wrong type of argument supplied to call.\n";
#ifdef    DEBUG

            *(Cor_exEnv->err_strm) << "Type of argument is: "
                << term.expr.kind_str() << endl;
#endif

            it.set_no_match();
            return(NULL);
        }

        Cor_CArg *rel_name = &(term.expr);
        rel = find_relation(rel_name->functor(), rel_name->arity());
        if (!rel) {
            *(Cor_exEnv->err_strm)
                 << "Relation " << rel_name->functor()
                 << " with arity " << rel_name->arity()
                 << " not found.\n";
            it.set_no_match();
            return(NULL);
        }

        // check arity as find_relation does not
        if ((rel->arity() > 0) && (rel->arity() != rel_name->arity())) {
            *(Cor_exEnv->err_strm) << "Relation arity does not match.\n";
            it.set_no_match();
            return(NULL);
        }

        newenv = new Cor_BindEnv(rel_name->max_vars());

        iter = new Cor_TupleIterator(rel, rel_name->args(), newenv);
        it.answer_iterator = iter;

    } else {
        it.stack_mark.pop_to();
        iter = it.answer_iterator;
        rel  = iter->relation;
    }

    Cor_BindEnv *result = iter->get_next();

    if (iter->no_match()) {
        it.set_no_match();
        return(NULL);
    } else {
        Cor_Term term(it.arg_list[0], env);
        Cor_deep_dereference_term(term);
        Cor_Term term2(term.expr, result);
        Cor_deep_dereference_term(term2);

        Cor_unify_args(term, term2);

        it.reset_no_match();
        return(result);
    }
}

// **************************************************************************
Cor_BindEnv *ArgTypeSolver(Cor_BuiltinRelation & rel,
                           Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (iterator.arg_list.count() != 1) {
        CORAL_error(COR_BAD_ARITY, "1 argument needed", NULL);
        iterator.set_no_match();
        return (iterator.bindenv);
    }
    Cor_Term term1(iterator.arg_list[0], iterator.bindenv);
    Cor_deep_dereference_term(term1);

    int (*type_function) (const Cor_CArg &);

    if (rel.name.equals(Cor_IsConstantSymbol))
        type_function = is_constant;
    else if (rel.name.equals(Cor_IsVarSymbol))
        type_function = is_var;
    else if (rel.name.equals(Cor_IsListSymbol))
        type_function = is_list;
    else if (rel.name.equals(Cor_IsFunctorSymbol))
        type_function = is_functor;
    else if (rel.name.equals(Cor_IsStringSymbol))
        type_function = is_string;
    else if (rel.name.equals(Cor_IsNumSymbol))
        type_function = is_num;
    else if (rel.name.equals(Cor_IsIntSymbol))
        type_function = is_int;
    else if (rel.name.equals(Cor_IsLongSymbol))
        type_function = is_long;
    else if (rel.name.equals(Cor_IsShortSymbol))
        type_function = is_short;
    else if (rel.name.equals(Cor_IsFloatSymbol))
        type_function = is_float;
    else if (rel.name.equals(Cor_IsDoubleSymbol))
        type_function = is_double;
    else {
        iterator.set_no_match();
        return (iterator.bindenv);
    }

    Cor_TermLink *rv = NULL;
    Cor_CArg tmp_carg;
    term1.expr.simplify(tmp_carg, iterator.bindenv, rv,
                        iterator.bindenv, NULL);

    if ((*type_function) (tmp_carg))
        iterator.reset_no_match();
    else
        iterator.set_no_match();

    return (iterator.bindenv);
}

// *************************************************************************
//  This function takes in 3 args. The 1st is base, 2nd is exp,
//  3rd is the answer to the query.
// *************************************************************************
Cor_Tuple *Cor_PowSolver(Cor_Tuple * in)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_ArgList *alist = new Cor_ArgList(3);
    Cor_CArg args[3];
    double base=0, exp=0, ans=0;
    Cor_CArg Ans;                // place for result.

    char type = 0;
    int numargs = 0;
    int numvars = 0;
    const int ROOT = 0, LOG = 1, POW = 2;

    if (in->arity() != 3) {
        CORAL_error(COR_BAD_ARITY,
             "Format of pow is: pow(Base,Exponent,Answer)", "PowSolver");
        return (NULL);
    }
    for (int i = 0; i < 3; i++) {
        args[i] = (in->args())[i];
        if (is_num(args[i])) {
            numargs++;
            if (is_int(args[i])) {
                if (i == ROOT)
                    base = (args[i].NumDval());
                else if (i == LOG)
                    exp = (args[i].NumDval());
                else if (i == POW)
                    ans = (args[i].NumDval());
            } else if (is_double(args[i])) {
                if (i == ROOT)
                    base = (args[i].NumDval());
                else if (i == LOG)
                    exp = (args[i].NumDval());
                else if (i == POW)
                    ans = (args[i].NumDval());
            } else {
                CORAL_error(COR_BAD_ARG_TYPE,
                       "Arg must be of type DOUBLE or INT", "PowSolver");
                return (NULL);
            }
        } else if (is_var(args[i])) {
            numvars++;
            type = i;
        } else {
            CORAL_error(COR_BAD_ARG_TYPE, "Arg must be a Variable or a Number",
                        "PowSolver");
            return (NULL);
        }
    }

    if (numargs == 2 && numvars == 1) {
        switch (type) {
        case POW:
            Ans = (pow(base, exp));        // call the function.

            (*alist)[0] = args[0];
            (*alist)[1] = args[1];
            (*alist)[2] = Ans;
            break;
        case LOG:
            Ans = ((log(ans) / log(base)));
            (*alist)[0] = args[0];
            (*alist)[1] = Ans;
            (*alist)[2] = args[2];
            break;
        case ROOT:
            Ans = (pow(ans, 1 / exp));        // call the function.

            (*alist)[0] = Ans;
            (*alist)[1] = args[1];
            (*alist)[2] = args[2];
            break;
        default:
            // error
            break;
        }
    } else if (numargs == 3 && numvars == 0) {        // still fine

        if (ans == pow(base, exp)) {
            (*alist)[0] = args[0];
            (*alist)[1] = args[1];
            (*alist)[2] = args[2];
        }
    } else {
        CORAL_error(COR_BAD_ARG_TYPE,
             "Format of pow is: pow(Base,Exponent,Answer)", "PowSolver");
        return (NULL);
    }
    Cor_Tuple *Result = new Cor_Tuple(alist);        // create the tuple.

    return (Result);
}

// *************************************************************************
Cor_Tuple *Cor_AbsSolver(Cor_Tuple *query)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (query == NULL) {
        *(Cor_exEnv->err_strm) << "Null argument passed to "
                               << Cor_AbsOpSymbol->symbol() << endl;
        return (NULL);
    }
    int arity;
    if ((arity = query->arity()) != 2) {
        CORAL_error(COR_BAD_ARITY, "2 arguments to abs needed", NULL);
        return (NULL);
    }
    if (!(query->args())[0].is_num()) {
        *(Cor_exEnv->err_strm)
              << "CORAL :: Error -- first argument to "
              << Cor_AbsOpSymbol->symbol() << " is not a number: ";
        (query->args())[0].printon(Cor_exEnv->err_strm);
        *(Cor_exEnv->err_strm) << endl;
        return (NULL);
    }
    if ((!(query->args())[1].is_num())
            && (!(query->args())[1].is_var())) {
        *(Cor_exEnv->err_strm)
              << "CORAL :: Error -- second argument to "
              << Cor_AbsOpSymbol->symbol()
              << " is neither free nor a number: ";
        (query->args())[1].printon(Cor_exEnv->err_strm);
        *(Cor_exEnv->err_strm) << endl;
        return (NULL);
    }
    Cor_ArgList *alist = new Cor_ArgList(arity);

    double arg0=0.0, arg1=0.0;
    int arg1_is_var = 0;

    (*alist)[0] = (query->args())[0];
    if (is_double((*alist)[0]))
        arg0 = ((*alist)[0]).NumDval();
    else if (is_int((*alist)[0]))
        arg0 = make_int((*alist)[0]);
    else
        return (NULL);

    if (is_double((query->args())[1]))
        arg1 = make_double((query->args())[1]);
    else if (is_int((query->args())[1]))
        arg1 = make_int((query->args())[1]);
    else
        arg1_is_var = 1;

    if (arg1_is_var) {
        (*alist)[1] = fabs(arg0);
    } else {
        if (fabs(arg0) == arg1)
            (*alist)[1] = (query->args())[1];
        else
            return (NULL);
    }

    Cor_Tuple *return_tuple = new Cor_Tuple(alist);
    return (return_tuple);
}

// *************************************************************************
Cor_Tuple *Cor_CeilSolver(Cor_Tuple *query)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    if (query == NULL) {
        *(Cor_exEnv->err_strm) << "Null argument passed to "
                               << Cor_CeilOpSymbol->symbol() << endl;
        return (NULL);
    }

    int arity;
    if ((arity = query->arity()) != 2) {
        CORAL_error(COR_BAD_ARITY, "2 arguments to ceiling needed", NULL);
        return (NULL);
    }

    if (!(query->args())[0].is_num()) {
        *(Cor_exEnv->err_strm)
              << "CORAL :: Error -- first argument to "
              << Cor_CeilOpSymbol->symbol() << " is not a number: ";
        (query->args())[0].printon(Cor_exEnv->err_strm);
        *(Cor_exEnv->err_strm) << endl;
        return (NULL);
    }

    if ((!(query->args())[1].is_num())
            && (!(query->args())[1].is_var())) {
        *(Cor_exEnv->err_strm)
              << "CORAL :: Error -- second argument to "
              << Cor_CeilOpSymbol->symbol()
              << " is neither free nor a number: ";
        (query->args())[1].printon(Cor_exEnv->err_strm);
        *(Cor_exEnv->err_strm) << endl;
        return (NULL);
    }
    Cor_ArgList *alist = new Cor_ArgList(arity);

    double arg0, arg1 = 0;
    int arg1_is_var = 0;

    (*alist)[0] = (query->args())[0];
    if (is_double((*alist)[0]))
        arg0 = make_double((*alist)[0]);
    else if (is_int((*alist)[0]))
        arg0 = make_int((*alist)[0]);
    else
        return (NULL);

    if (is_double((query->args())[1]))
        arg1 = make_double((query->args())[1]);
    else if (is_int((query->args())[1]))
        arg1 = make_int((query->args())[1]);
    else
        arg1_is_var = 1;

    if (arg1_is_var) {
        (*alist)[1] = (ceil(arg0));
    } else {
        if (ceil(arg0) == arg1)
            (*alist)[1] = (query->args())[1];
        else
            return (NULL);
    }

    Cor_Tuple *return_tuple = new Cor_Tuple(alist);
    return (return_tuple);
}

// *************************************************************************
Cor_Tuple *Cor_FloorSolver(Cor_Tuple *query)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (query == NULL) {
        *(Cor_exEnv->err_strm) << "Null argument passed to "
                               << Cor_FloorOpSymbol->symbol() << endl;
        return (NULL);
    }

    switch (Cor_CheckArgTypes(query->args(), 2, 2, "nv")) {

    case Coral_Not_Enough_Args:
    case Coral_Too_Many_Args:
        CORAL_error(COR_BAD_ARITY, "2 arguments to floor needed", NULL);
        return (NULL);

    case Coral_Args_Okay:
        break;

    case Coral_Wrong_Type_Args:
    default:
        if ((query->args())[0].is_num() && (query->args())[1].is_num())
            break;

        *(Cor_exEnv->err_strm) << "Error: " << *Cor_FloorOpSymbol
            << "() accepts forms bb or bf with bound arguments numeric.\n";
        break;
    }

    Cor_ArgList *alist = new Cor_ArgList(query->arity());

    double arg0, arg1=0;
    int arg1_is_var = 0;

    (*alist)[0] = (query->args())[0];
    if (is_double((*alist)[0]))
        arg0 = make_double((*alist)[0]);
    else if (is_int((*alist)[0]))
        arg0 = make_int((*alist)[0]);
    else
        return(NULL);

    if (is_double((query->args())[1]))
        arg1 = make_double((query->args())[1]);
    else if (is_int((query->args())[1]))
        arg1 = make_int((query->args())[1]);
    else
        arg1_is_var = 1;

    if (arg1_is_var) {
        (*alist)[1] = floor(arg0);
    } else {
        if (floor(arg0) == arg1)
            (*alist)[1] = (query->args())[1];
        else
            return(NULL);
    }

    Cor_Tuple *return_tuple = new Cor_Tuple(alist);
    return(return_tuple);
}

// *************************************************************************
Cor_Tuple *Cor_ExtractSolver(Cor_Tuple *query)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (query == NULL) {
        *(Cor_exEnv->err_strm) << "Null argument passed to ExtractSolver\n";
        return (NULL);
    }

    int arity = query->arity();
    Cor_ArgList *alist = new Cor_ArgList(arity);
    (*alist)[0] = (query->args())[0];
    (*alist)[1] = (query->args())[1];

    char *arg0 = (char *) make_string((query->args())[0]);
    int arg1 = (int) make_int((query->args())[1]);
    char *retvalue;
    if ((unsigned int) arg1 >= strlen(arg0)) {
        delete [] arg0;
        delete alist;
        return (NULL);
    }
    retvalue = new char[2];
    retvalue[0] = arg0[arg1];
    retvalue[1] = '\0';

    Cor_CArg *carg = make_arg(retvalue);
    (*alist)[2] = *carg;
    carg->destroy();
    delete [] retvalue;

    Cor_Tuple *return_tuple = new Cor_Tuple(alist);
    return (return_tuple);
}

// *************************************************************************
Cor_Tuple *Cor_LengthSolver(Cor_Tuple *query)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int retvalue;

    if (query == NULL) {
        *(Cor_exEnv->err_strm) << "Null argument passed to LengthSolver\n";
        return(NULL);
    }

    int arity = query->arity();
    Cor_ArgList *alist = new Cor_ArgList(arity);

    (*alist)[0] = (query->args())[0];

    if ( (*alist)[0].is_list() ) {
        retvalue = (*alist)[0].num_items();
        if (retvalue)
            retvalue--;
    } else {
        char *arg0 = (char *) make_string((query->args())[0]);
        retvalue = strlen(arg0);
    }

    (*alist)[1] = retvalue;
    Cor_Tuple *return_tuple = new Cor_Tuple(alist);

    return(return_tuple);
}

// *************************************************************************
Cor_Tuple *Cor_SubstringSolver(Cor_Tuple *query)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (query == NULL) {
        *(Cor_exEnv->err_strm) <<
                            "Null argument passed to SubstringSolver\n";
        return (NULL);
    }
    int arity = query->arity();
    Cor_ArgList *alist = new Cor_ArgList(arity);
    (*alist)[0] = (query->args())[0];
    (*alist)[1] = (query->args())[1];

    char *arg0;
    arg0 = (char *) make_string((query->args())[0]);
    char *arg1;
    arg1 = (char *) make_string((query->args())[1]);

    char *tmpstr = strstr(arg0, arg1);
    if (!tmpstr)
        tmpstr = "";

    Cor_CArg *carg = make_arg(tmpstr);
    (*alist)[2] = *carg;
    carg->destroy();

    Cor_Tuple *return_tuple = new Cor_Tuple(alist);
    delete [] arg0;
    delete [] arg1;
    return (return_tuple);
}

// *************************************************************************
#define FAIL_RETURN  {iterator.set_no_match(); return(NULL);}

Cor_BindEnv *Cor_RelOptionsSolver(Cor_BuiltinRelation & rel,
                                  Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (iterator.arg_list.count() == 0) {
        if (rel.name.equals(Cor_ExplainOnSymbol)) {
            Cor_exEnv->GlobalRelationOptions |= REL_EXPLAIN;
            if (coralConfig->server_mode && (!Cor_curUser->explain_strm)) {

                Cor_sockbuf *expl_buf = (Cor_sockbuf *)
                Cor_clients->expl_serv_strm->rdbuf();

#ifdef    DEBUG
                cerr << endl
                     << CP_EXPL << Cor_curConn->conn_id << " "
                     << Cor_curUser->user_id << " "
                     << expl_buf->hostname() << " "
                     << expl_buf->my_port() << endl;
#endif
                // Tell client to start explain
                *(Cor_exEnv->out_strm) << endl
                     << CP_EXPL << Cor_curConn->conn_id << " "
                     << Cor_curUser->user_id << " "
                     << expl_buf->hostname() << " "
                     << expl_buf->my_port() << endl << flush;

                // Client will block until Explain connects
                Cor_curUser->yield();
            }
        } else if (rel.name.equals(Cor_ExplainOffSymbol)) {
            Cor_exEnv->GlobalRelationOptions &= ~REL_EXPLAIN;
        } else if (rel.name.equals(Cor_TraceSymbol)) {
            Cor_exEnv->GlobalRelationOptions |= REL_DISPLAY_INSERTIONS;
        } else if (rel.name.equals(Cor_UntraceSymbol)) {
            Cor_exEnv->GlobalRelationOptions &= ~REL_DISPLAY_INSERTIONS;
        } else if (rel.name.equals(Cor_ProfileOnSymbol)) {
            Cor_exEnv->GlobalRelationOptions |= REL_PROFILE_STATS;
        } else if (rel.name.equals(Cor_ProfileOffSymbol)) {
            Cor_exEnv->GlobalRelationOptions &= ~REL_PROFILE_STATS;
        }
        return (iterator.bindenv);
    }
    FOR_EACH_ARG(arg, iterator.arg_list) {

        Cor_Term term1(*arg, iterator.bindenv);
        Cor_deep_dereference_term(term1);
        if (!term1.expr.is_functor()) {
            term1.expr.printon(Cor_exEnv->err_strm);
            CORAL_error(COR_BAD_ARGLIST, NULL, "RelOptionsSolver");
            FAIL_RETURN;
        }
        Cor_CArg *fa = &(term1.expr);
        if ((fa->arity() != 1) || !fa->args()[0].is_symbol()) {
            term1.expr.printon(Cor_exEnv->err_strm);
            CORAL_error(COR_BAD_ARGLIST, NULL, "RelOptionsSolver");
            FAIL_RETURN;
        }
        char *adorn = fa->args()[0].symbol();

        Cor_CArg tmp = Cor_CArg(fa->functor());
        Cor_Relation *newrel =
            Cor_find_external_relation(tmp, strlen(adorn));

        if (!newrel) {
            CORAL_error(COR_BAD_REL_NAME, term1.expr.symbol(),
                        "RelOptionsSolver");
            FAIL_RETURN;
        }
        if (rel.name.equals(Cor_TraceSymbol)) {
            Cor_orOptionToRelation(newrel, adorn, REL_DISPLAY_INSERTIONS);
        } else if (rel.name.equals(Cor_UntraceSymbol)) {
            Cor_andOptionToRelation(newrel, adorn, ~REL_DISPLAY_INSERTIONS);
        } else if (rel.name.equals(Cor_ProfileOnSymbol)) {
            Cor_orOptionToRelation(newrel, adorn, REL_PROFILE_STATS);
        } else if (rel.name.equals(Cor_ProfileOffSymbol)) {
            Cor_andOptionToRelation(newrel, adorn, ~REL_PROFILE_STATS);
        } else if (rel.name.equals(Cor_ExplainOnSymbol)) {
            Cor_orOptionToRelation(newrel, adorn, REL_EXPLAIN);
        } else if (rel.name.equals(Cor_ExplainOffSymbol)) {
            Cor_andOptionToRelation(newrel, adorn, ~REL_EXPLAIN);
        }
    }
    END_EACH_ARG;

    return(iterator.bindenv);
}

#undef FAIL_RETURN

// *************************************************************************
Cor_BindEnv *Cor_SortRelSolver(Cor_BuiltinRelation&,
                Cor_TupleIterator& it)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_ArgType_Error err;
    Cor_Term terms[4];
    int  incr_order = 1;

    err = Cor_CheckArgTypes(it.arg_list, 3, 4, "flss", terms, 1, it.bindenv);

    if (err != Coral_Args_Okay) {
        *(Cor_exEnv->err_strm) << "Error: " << *Cor_SortSymbol
            << "(): Invalid arguments.\n";
        it.set_no_match();
        return(NULL);
    }

    if (it.arg_list.count() == 4) {
        if (Cor_DecrSymbol->equals(terms[3].expr))
            incr_order = 0;
        else if (Cor_IncrSymbol->equals(terms[3].expr))
            incr_order = 1;
        else
            *(Cor_exEnv->err_strm) << "Error: " << *Cor_SortSymbol
                << "(): Invalid ordering, only " << *Cor_IncrSymbol
                << " and " << *Cor_DecrSymbol << " allowed.\n";
    }

    int rel_arity = terms[0].expr.arity();
    Cor_CArg *relnme = terms[0].expr.cfunctor();
    Cor_Relation *inrel, *chkrel;

    inrel = Cor_find_external_relation(*relnme, rel_arity);

    if (!inrel) {
        *(Cor_exEnv->err_strm) << "Error: " << *Cor_SortSymbol
            << "(): Relation " << *relnme << "/" << rel_arity
            << " not found.\n";
        it.set_no_match();
        return(NULL);
    }

    chkrel = Cor_find_external_relation(terms[2].expr, rel_arity);

    if (chkrel) {
        *(Cor_exEnv->err_strm) << "Error: " << *Cor_SortSymbol
            << "(): Relation " << terms[2].expr << "/" << rel_arity
            << " must not already exist.\n";
        it.set_no_match();
        return(NULL);
    }

    Cor_IntStack *order = new Cor_IntStack();
    Cor_CArgStack *ord_list = terms[1].expr.items();

    int j, k, cnt;
    cnt = ord_list->count();

    for (j=cnt-1; j > 0; j--) {

        Cor_CArg *item = ord_list->get_ith(j);

        if (item->is_var()) {
            for (k=0; k < rel_arity; k++)
                if (item->equals( (terms[0].expr.args())[k]) ) {
                    order->push(k);
                    break;
                }
        } else if (item->is_int() && (item->NumIval() < rel_arity)) {
            order->push(item->NumIval());
        } else
            *(Cor_exEnv->err_strm) << "Warning: " << *Cor_SortSymbol
                << "(): Invalid sort order argument, '"
                << *item << "' ignored.\n";
    }

    Cor_SortedRelation *sorted =
            new Cor_SortedRelation(rel_arity,order,incr_order);
    sorted->name = terms[2].expr;
    Cor_CurDB->RelationTable->insert(sorted->name, sorted);

    Cor_StackMark stack_mark0;
    Cor_BindEnv env0(rel_arity);
    Cor_TupleIterator iter(inrel, terms[0].expr.args(), &env0);

    for (;;) {
        Cor_Tuple *tup = iter.get_next_tuple();

        if (iter.no_match())
            break;

        Cor_Tuple *new_tup = new Cor_Tuple(tup->args().dup());

        if (! sorted->insert_tuple(new_tup))
            delete new_tup;
    }

    stack_mark0.pop_to();

    it.reset_no_match();
    return(it.bindenv);
}

// *************************************************************************
