// *************************************************************************
//    SymTable.h
//
//    Shaun Flisakowski - May 1995
// 
//    This Symbol Table is a replacement for an older one, it needed
//    to have changes made to it; but it had been so throughly 
//    hand-optimized that it was no longer understandable. 
//
//    There is a more generic hash table in the hashtable directory, this
//    is just a very simple one for mapping a Cor_CArg to a void *.
//
// *************************************************************************

#ifndef _SYMTABLE_H_
#define _SYMTABLE_H_

#include "CArg.h"


// *************************************************************************

    // What is a reasonable size for this? (it can grow)
#define    COR_DEFLT_SYMTAB_SIZE    31

// *************************************************************************
//  For a pointer to whatever

typedef  void *Pointer;

// *************************************************************************

class Cor_SymTable;

class Cor_SymTabElement {

    friend class Cor_SymTable;
    friend class Cor_SymCursor;

private:
        // For linking in buckets
    Cor_SymTabElement *_next;

    void set_next(Cor_SymTabElement *next) { _next = next; }
    Cor_SymTabElement *get_next() { return(_next); }

public:

    Pointer    data;    
    Cor_CArg   arg;

    Cor_SymTabElement();
    Cor_SymTabElement(Cor_CArg& ste_arg, Pointer ste_data);
   ~Cor_SymTabElement();
};

// *************************************************************************

class Cor_SymCursor {

    friend class Cor_SymTable;

private:

    int index;                      // The element of the array currently
                                    // being examined
    Cor_SymTabElement *current;     // NULL indicates to move to next index

    Cor_SymTable      *table;       // The symbol table that is
                                    // being scanned 
public:

    Cor_SymCursor(Cor_SymTable *table);
   ~Cor_SymCursor();

        // To move to the next item (False when done)
    int ScanNext();

        // To get the data out
    Cor_CArg& arg() { return(current->arg); }
    Pointer data() { return( (current) ? (current->data) : NULL ); }
};

// *************************************************************************

class Cor_SymTable {

    friend class Cor_SymCursor;

private:
    int     _size;                // The size of the table
    int     _num_elements;        // The number of elements currently stored

    int     _element_limit;       // Expand table if the number of elements
                                  // grows larger than this (Currently
                                  //  expanded via doubling size
                                  //  and rehashing all elements)

    Cor_SymTabElement **_symtab;  // The table itself

public:

        // size of zero uses COR_DEFLT_SYMTAB_SIZE
    Cor_SymTable(int size = 0 );
   ~Cor_SymTable();

    int size() { return(_size); }
    int num_elements() { return(_num_elements); }

        // Warning, this will insert duplicates, data cannot be NULL
    int  insert(Cor_CArg& arg, Pointer data);

        // This one checks for duplicates before inserting, data cannot be NULL
    Pointer insert_nodup(Cor_CArg& arg, Pointer data);

        // Returns a Pointer to data if delete found its target
    Pointer delete_entry(Cor_CArg& arg);

        // Returns NULL if arg is not in table, otherwise the data.
    Pointer lookup(Cor_CArg& arg);

        // Delete all items in the table
    void clear();

        // Grow the hash table larger (double it)
    void rehash_bigger();

        // For scanning the entire table
    Cor_SymCursor *OpenScan();
};

// *************************************************************************

#endif _SYMTABLE_H_
