// *************************************************************************
//    SupMagic.C
// *************************************************************************

#include <stdio.h>
#include <string.h>

#include "SupMagic.h"

#include "Rule.h"
#include "Interface.h"
#include "Globals.h"
#include "Unparse.h"
#include "CompileRule.h"
#include "BuiltinSymbols.h"
#include "Literal.h"
#include "ExportInfo.h"

// *************************************************************************

#ifndef MAX_NAME_LENGTH
#define MAX_NAME_LENGTH (256+1)
#endif

// *************************************************************************

extern int opt_share_functors;

// *************************************************************************

static Cor_Literal *build_pred(Cor_Clause *adrule, Cor_Literal *pred,
        int argnum);

static Cor_IntStack *vars_from(Cor_Clause *adrule, int pred);
static Cor_Literal *build_supmagic(Cor_Clause *adrule, int rule_num, int pos,
    Cor_IntStack *var_list, int next_der_lit);

// *************************************************************************

static int dont_create_sup0 = 1;

static int rule_num = 0;

// *************************************************************************
/*
    This function does an enhanced form of supplementary magic, where each
    rule can have several body literals, but only the first two literals
    can have derived predicates (i.e. predicates defined in the given module.  
    Semi-naive rewriting of such rules can safely permute these two derived
    literals if required, without causing any join order problems.
*/

void supmagic(Cor_Collection *ad_rules, Cor_Collection *magic_rules)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    rule_num = 0;
    Cor_Clause *adrule;
    int j, nrules;

    nrules = ad_rules->count();

    for (j=0; j < nrules; j++) {
        adrule = ad_rules->get_ith(j);

          // Construct the supplementary magic rules.
        if (0 && Cor_parserStruct->CurModule->NestedSccEval && !adrule->supp)
          continue;

        do_supp_magic_on_rule(adrule, magic_rules);
    }
}

// *************************************************************************
// Construct the supplementary magic rules for adrule

void do_supp_magic_on_rule(Cor_Clause *adrule, Cor_Collection *magic_rules)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int    i, j, num_base_lits;
    int    tmp_j, inc;
    Cor_Clause *magicrule;
    Cor_Literal *pred;
    Cor_Literal *newhead;
    Cor_Literal *prev_sup;

    rule_num++;
    newhead = build_pred(adrule, adrule->head,0);

    for (j=0;;) {
        if (j < adrule->num_preds && adrule->preds[j]->base_pred)
            j++;
        else
            break;
    }
    num_base_lits = j;
    if (num_base_lits < adrule->num_preds && dont_create_sup0)
     {
        // Avoid generating sup_j_0 predicate - in most cases this is an
        // unnecessary overhead, but is necessary when using supplementary
        // indexing or return unify.
        prev_sup = build_magic_pred(adrule, adrule->head, 0, 0);
        num_base_lits = 0; 
     } else {
        // Create sup_i_0 predicate.
        magicrule = Cor_copy_rule(adrule);
            Cor_alloc_preds(magicrule, num_base_lits+1);
        magicrule->preds[0] = build_magic_pred(adrule, adrule->head, 0, 0);
        for (tmp_j=0;tmp_j < num_base_lits; tmp_j++) {
              magicrule->preds[tmp_j+1] = adrule->preds[tmp_j];
        }

        if (num_base_lits >= adrule->num_preds) {
            pred = newhead;
            magicrule->head_deletes_count = adrule->head_deletes_count;
            magicrule->head_deletes = adrule->head_deletes;
        } else {
            // Keep only those vars that either appear in the head or later
            // in the original rule AND appear in the body of the
            // supplementary rule that we are going to generate.
            Cor_IntStack *supvars = new Cor_IntStack();
            supvars = Cor_build_var_list(magicrule->preds[0], 
                  Cor_vars_between(adrule, 0, num_base_lits-1, supvars, NULL));
            Cor_IntStack *latervars = Cor_build_var_list(newhead,
                                        vars_from(adrule,num_base_lits));
            supvars = Cor_intersect_var_lists(supvars, latervars);
            pred = build_supmagic(adrule, rule_num, 0, supvars, num_base_lits);
        }

        // WARNING - unnecessary vars should be projected out above
        // Perhaps all vars are necessary in this case though.
        magicrule->head = prev_sup = pred;
        magic_rules->push(magicrule);
        pred->rule = adrule;
    }
  
    for (i=num_base_lits; i < adrule->num_preds;) {
        // Generate a magic rule for each derived right-hand side predicate
        // that has an incoming arc in the sip.

        if ( 0 && ! adrule->preds[i]->base_pred && adrule->sips->sip_list ) {
            // Don't use this for now
            // WARNING: Eventually just sort the predicates in the sip order,
            // and then do the sup magic is below.
            if (0) {
            *cerr_strm << "Error: supmagic cannot deal with sips yet\n";
            Cor_Sip *s;
            for (s=adrule->sips->sip_list; s; s=s->next)
              if (i == s->head-1) {
                magicrule = Cor_copy_rule(adrule);
                // rule_num++;
                pred = build_magic_pred(adrule, adrule->preds[i],
                    0,                  // use_share_opts
                    i + 1);             // argnum 
                magicrule->head = pred;
                Cor_alloc_preds(magicrule, 1);
                pred->rule = adrule;
                pred = build_supmagic(adrule, rule_num, i+1,
                      Cor_build_var_list(prev_sup, NULL),
                      /* derived lit pos.??*/     i);
                magicrule->preds[0] = pred;
            
                magicrule->type = COR_MAGICRULE;
                magicrule->source_rule = adrule;
                magicrule->source_pred = adrule->preds[i];
                magic_rules->push(magicrule);
            
                break;    /* only one magic rule per predicate */
            }
          }
        }
    
        if ( !(adrule->preds[i]->base_pred))  {
            magicrule = Cor_copy_rule(adrule);
            magicrule->head = build_magic_pred(adrule, adrule->preds[i],0,i+1);
                  Cor_alloc_preds(magicrule, 1);
                  magicrule->preds[0] = prev_sup;
                magicrule->head->rule = adrule;
            magicrule->type = COR_MAGICRULE;
            magicrule->source_rule = adrule;
            magicrule->source_pred = adrule->preds[i];

            if (magicrule->source_pred->negated ||
                    adrule->head->args.has_grouping_args()) {

                if (magicrule->clause_names->is_empty()) {
                      // need to create a new name
                    magicrule->clause_names->add(*Cor_DontCareSymbol,1);
                }
                magicrule->clause_names->set_negated(0,1);
            }
            magic_rules->push(magicrule);
        }

        // Generate supplementary rule/ final rule.
        for (j=0;;) {
            if (i+j+1 < adrule->num_preds && adrule->preds[i+j+1]->base_pred)
                j++;
            else
                break;
        }
        num_base_lits = j;

        magicrule = Cor_copy_rule(adrule);

        if (i+num_base_lits+1 < adrule->num_preds) {
            // Keep only those vars that either appear in the head or later in
            // the original rule AND appear in the body of the supplementary
            // rule that we are going to generate.
            Cor_IntStack *supvars = new Cor_IntStack();
            supvars = Cor_build_var_list(prev_sup,
                Cor_vars_between(adrule, i, i+num_base_lits, supvars,NULL));
            Cor_IntStack *latervars = Cor_build_var_list(newhead,
                vars_from(adrule,i+num_base_lits+1));
            supvars = Cor_intersect_var_lists(supvars, latervars);
            pred = build_supmagic(adrule, rule_num, i+1,
                                        supvars,i+num_base_lits+1);
        } else {
            pred = newhead;
            magicrule->head_deletes_count = adrule->head_deletes_count;
            magicrule->head_deletes = adrule->head_deletes;
        }

        int num_deriv_negated = 0;
        if (Cor_parserStruct->CurModule->UseOrdSearch) {
            if ((adrule->preds[i]->negated || ((pred == newhead) &&
                newhead->args.has_grouping_args())) && 
                !Cor_parserStruct->CurModule->Monotonic &&
                !adrule->preds[i]->base_pred)
            {
                num_deriv_negated = 1;
            }
        }

        magicrule->head = pred;
        pred->rule = adrule;

        // added by Ken
        if (0 && Cor_parserStruct->CurModule->NestedSccEval &&
            prev_sup->kind == COR_SUPMAGIC &&
            !prev_sup->predicate->ext_magic)
          inc = prev_sup->supp_rule->num_preds - 1;
        else
          inc = 0;
            Cor_alloc_preds(magicrule, inc+2+num_base_lits+num_deriv_negated);
        magicrule->preds[0] = prev_sup;

        if (0 && Cor_parserStruct->CurModule->NestedSccEval &&
            prev_sup->kind == COR_SUPMAGIC &&
            !prev_sup->predicate->ext_magic)
          // copy the body of the rule of the prev_sup to this rule
          for (int pos = 0; pos < prev_sup->supp_rule->num_preds; pos++)
            magicrule->preds[pos] = prev_sup->supp_rule->preds[pos];

        if (num_deriv_negated) {
            magicrule->preds[inc+1] =
                build_magic_pred(adrule,adrule->preds[i],0,i+1);
            magicrule->preds[inc+1]->kind =  COR_DONE_MAGIC;
        }

        magicrule->preds[inc+num_deriv_negated+1] =
            build_pred(adrule, adrule->preds[i], i+1);

        for (tmp_j=0;tmp_j<num_base_lits; tmp_j++) {
            magicrule->preds[inc+tmp_j+num_deriv_negated+2] =
                adrule->preds[i+tmp_j+1];
        }
        i=i+num_base_lits+1;
        
        // If in nested_scc mode, mark the predicate if it will 
        // actually be used in the body of other rules. Added by Ken
        if (0 && Cor_parserStruct->CurModule->NestedSccEval &&
            i < adrule->num_preds &&
            pred->kind == COR_SUPMAGIC &&
            adrule->preds[i]->predicate->scc->scc_no !=
            adrule->head->predicate->scc->scc_no)
          // the next derived predicate is not in the same Scc as head
          // this sup predicate will be used in the body of other rule
          pred->predicate->ext_magic = 1;
          
        pred->supp_rule = magicrule;
        prev_sup = pred;
        if (1 || !Cor_parserStruct->CurModule->NestedSccEval ||
            pred->predicate->ext_magic == 1 || 
            pred->kind != COR_SUPMAGIC)
          magic_rules->push(magicrule);
        //generate_index_annotations(magicrule);
        // print_rule(magicrule,cerr_strm);

        /* WARNING::  Create make_index annotation here!! */
    }
}

// *************************************************************************
/*
    This procedure adds new variable names to rule, and stores the number
    of vars at the time it was called in old_num_var_names.
*/

void make_new_rule_vars(Cor_Clause *, int)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *cerr_strm << "make_new_rule_vars not implemented fully yet\n";
}

// *************************************************************************
/*
    This procedure returns the newvarnum'th new rule variable.  The new
    variables must have been created by make_new_rule_vars.
*/

Cor_CArg *new_rule_var(Cor_Clause *, int)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *cerr_strm << "new_rule_var not implemented fully yet\n";
    return NULL;
}

// *************************************************************************
Cor_CArg *make_supindex_var(Cor_Clause *adrule, int varnum, char letter)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char n[30];
    int varpos;

    sprintf(n,"_%c%d", letter, varnum);
    Cor_CArg *carg = new Cor_CArg(n);
    int rule_cnt;

    if (adrule->var_names)
        rule_cnt = adrule->var_names->count();
    else
        rule_cnt = 0;

    for (varpos = 0; varpos < rule_cnt; varpos++) {
        if ((adrule->var_names->get_ith(varpos))->nm_equals(*carg))
            break;
    }

    adrule->var_names->push(carg);
    ASSERT(varpos == rule_cnt);

    carg->set_var(varpos);
    return(carg);
}

// *************************************************************************
Cor_CArg& new_pred_name(Cor_CArg& name) 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return name;
}

// *************************************************************************
Cor_CArg *new_pred_name(char *name) 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return(new Cor_CArg(name));
}

// *************************************************************************
Cor_CArg *new_interface_pred_name(char *name) 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return(new Cor_CArg(name));
}

// *************************************************************************
// Construct an arg list containing only the bound args from the given pred.

Cor_Literal *build_magic_pred(Cor_Clause *adrule, Cor_Literal * pred,
    int use_share_opt, int /* argnum */)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int    i;
    Cor_CArg *arg_ptr;
    Cor_Literal *new_pred;
    int count = 0; // count of bound args in pred
    int new_rule_var_count = 0;

    i = 0;
    FOR_EACH_ARG(a, pred->args) {
        if (pred->adorn.test(i++))
            count++;
    } END_EACH_ARG

    new_pred = (Cor_Literal*) new Cor_Literal(count);
    new_pred->kind = COR_MAGIC;
    arg_ptr = &(new_pred->args[0]);
    i = 0;
    FOR_EACH_ARG(a, pred->args) {
        if (pred->adorn.test(i++)) {
            if (!use_share_opt || ! a->is_functor())
                *arg_ptr++ = *a;
            else
                *arg_ptr++ = *(new_rule_var(adrule, new_rule_var_count++));
        }
    } END_EACH_ARG
    new_pred->base_pred = 0;
    new_pred->negated = 0;
    new_pred->pred = pred->pred;
    new_pred->adorn = pred->adorn;
    return new_pred;
}

// *************************************************************************
Cor_Literal *build_interface_magic_pred(Cor_Clause *adrule, Cor_Literal *pred,
    int use_share_opt, int argnum)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Literal *mlit = build_magic_pred(adrule, pred, use_share_opt, argnum);
    Cor_CArg *tmp = new_interface_pred_name(pred->pred.symbol());
    mlit->pred = *tmp;
    delete tmp;
    mlit->kind = COR_REGULAR;
    // So that m_ is not printed before predicate name
    return mlit;
}

// *************************************************************************
/*
//  Create a new predicate from pred.  The new predicate will have
//  an extra argument if supplementary indexing rewriting is used.
//  Create a variable using number ``argnum'' and use it in this 
//  extra argument.
*/

static Cor_Literal *build_pred(Cor_Clause* /* adrule */, Cor_Literal *pred,
    int /* argnum */)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return(pred);
}

// *************************************************************************
int count_bound_functor_args(Cor_Literal *head) 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int count = 0;
    int i = 0;

    FOR_EACH_ARG(a,head->args) {
        if (head->adorn.test(i++) && a->is_functor())
            count++;
    } END_EACH_ARG
    
    return(count);
}

// *************************************************************************
/*
    The following fn is used in the opt_share_functors optimization.
    It creates a literal of the form "Var=term", where the var is 
    specified via newvarnum, and "term" is the argnum'th arg of head.
*/

Cor_Literal *make_functor_literal(Cor_Clause *newrule,
        Cor_Literal *head, int argnum, int newvarnum)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *cerr_strm << "make_functor_literal not implemented fully yet\n";

    Cor_Literal *new_pred; 
    Cor_CArg *arg_ptr;

    new_pred = (Cor_Literal*)
       malloc(sizeof(Cor_Literal) + 2 * sizeof(Cor_CArg *));
    new_pred->kind = COR_REGULAR;
    new_pred->args.set_count(2);
    arg_ptr = &(new_pred->args[0]);

    *arg_ptr++ = *(new_rule_var(newrule, newvarnum));
    *arg_ptr = head->args[argnum];

    new_pred->base_pred = 1;
    new_pred->negated = 0;
    new_pred->pred = *Cor_EquOpSymbol;
    // new_pred->adorn is not initialized
    return new_pred;
}


// *************************************************************************
//
//  This procedure adds literals of the form var=term, for the
//  opt_share_functors optimization
//
void create_functor_literals(Cor_Clause *)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *cerr_strm << "create_functor_literals not implemented yet\n";
}

// *************************************************************************
/*
    This procedure creates a head literal for the magic rule, and implements
    the opt_share_functors optimization if the global flag opt_share_functors
    is set.
*/

Cor_Literal *build_head_pred(Cor_Clause *newrule, Cor_Literal *oldhead)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int new_rule_var_count = 0;
    int i=0;
    Cor_Literal *new_pred; 
    Cor_CArg *arg_ptr;

    if (!opt_share_functors || count_bound_functor_args(oldhead) == 0)
        return(oldhead);

    *cerr_strm << "build_head_pred not implemented fully yet\n";

    new_pred = (Cor_Literal*)
        malloc(sizeof(Cor_Literal) + oldhead->arity() * sizeof(Cor_CArg *));
    new_pred->kind = COR_REGULAR;
    new_pred->args.set_count(oldhead->arity());
    arg_ptr = &(new_pred->args[0]);

    FOR_EACH_ARG(a, oldhead->args) {
        if ( oldhead->adorn.test(i++) && opt_share_functors &&
            a->is_functor())
        {
            *arg_ptr++ = *(new_rule_var(newrule, new_rule_var_count++));
        }
        else {
            *arg_ptr++ = *a;
        }
    } END_EACH_ARG
    new_pred->base_pred = 0;
    new_pred->negated = 0;
    new_pred->pred = oldhead->pred;
    new_pred->adorn = oldhead->adorn;
    return new_pred;
}

// *************************************************************************
static Cor_IntStack *vars_from(Cor_Clause *adrule, int pred)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;
    Cor_IntStack *result = new Cor_IntStack();
    for (i=pred; i < adrule->num_preds; i++)
        Cor_build_var_list(adrule->preds[i], result);
    return result;
}

// *************************************************************************
// WARNING: The following routine damages the varlists that are passed to it.

static Cor_Literal *build_supmagic(Cor_Clause* /* adrule*/, int rule_num,
    int pos, Cor_IntStack *var_list, int /* next_der_lit */)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Literal *pred;
    Cor_CArg    *arg_ptr;

    int var_count = var_list->count();

    static char name[30];
    sprintf(name, "sup_%d_%d", rule_num, pos);
    Cor_CArg tmp_arg = Cor_CArg(name);
    pred = new Cor_Literal(Cor_parserStruct, tmp_arg, var_count);
    // Used to be pred = new Literal(NULL, var_count);
    pred->kind = COR_SUPMAGIC;
    pred->pos = pos;
    arg_ptr = &pred->args[var_count];

    Cor_CArg varg;
    while (! var_list->is_empty()) {
        varg.set_var(var_list->pop());
        *--arg_ptr = varg;
    }
    delete var_list;

    return pred;
}

// *************************************************************************
void create_interface_index(Cor_Literal *pred)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int arity = pred->args.count();

    char *buf = new char[arity+1];
    buf[0] = 'b';
    for (int i=1; i < arity; i++)
        buf[i] = 'f';
    buf[arity] = '\0';
    
    char *name = new char [strlen(pred->pred.symbol())+10];
    Cor_sprint_pred_name(name, pred->pred.symbol(),"");

    Cor_CArg tmp_arg = Cor_CArg(name);
    Cor_add_make_index_annotation(Cor_parserStruct->make_index_annotations,
        tmp_arg, &(pred->adorn),make_arglist(1,make_arg(buf)), NULL);

    delete [] buf;
    delete [] name;
}

// *************************************************************************
Cor_Literal *make_vararglist_literal(Cor_Clause *rule, Cor_CArg& name,
        int startvarnum, int arity)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;
    Cor_Literal *lit = new Cor_Literal(Cor_parserStruct, name, arity);
    for(i=0; i < arity; i++) {
        lit->args[i] = *(make_supindex_var(rule, i+startvarnum, 'V'));
    }
    return lit;
}

// *************************************************************************
Cor_Literal *make_vararglist_literal(Cor_Clause *rule, char *name,
    int startvarnum, int arity)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg tmp_arg = Cor_CArg(name);
    return make_vararglist_literal(rule, tmp_arg, startvarnum, arity);
}

// *************************************************************************
